##
## This file is part of the exputils package.
##
## Copyright: INRIA
## Year: 2022
## Contact: chris.reinke@inria.fr
##
## exputils is provided under GPL-3.0-or-later
##
import os

def makedirs(path):
    '''
    Creates the directories of the given path if they do not exist.

    :param path:  Directory path.
    :return:
    '''

    if not os.path.isdir(path):
        os.makedirs(path)


def makedirs_for_file(filepath):
    '''
    Creates the directory of the given filepath if it does not exist.

    :param filepath: Filepath for which the directories it points to are created.
    :return:
    '''

    directory_path, _ = os.path.split(filepath)
    makedirs(directory_path)




##
## This file is part of the exputils package.
##
## Copyright: INRIA
## Year: 2022
## Contact: chris.reinke@inria.fr
##
## exputils is provided under GPL-3.0-or-later
##
from exputils.io.numpy import load_numpy_files
from exputils.io.numpy import save_dict_to_numpy_files

from exputils.io.general import makedirs
from exputils.io.general import makedirs_for_file

from exputils.io.odsreader import ODSReader

from exputils.io.json import ExputilsJSONEncoder
from exputils.io.json import exputils_json_object_hook
from exputils.io.json import save_dict_as_json_file
from exputils.io.json import load_dict_from_json_file
from exputils.io.json import convert_json_dict_keys_to_ints

from exputils.io.dill import load_dill
from exputils.io.dill import save_dill
from exputils.io.dill import load_dill_files



##
## This file is part of the exputils package.
##
## Copyright: INRIA
## Year: 2022
## Contact: chris.reinke@inria.fr
##
## exputils is provided under GPL-3.0-or-later
##
import exputils as eu
import numpy as np
import scipy.stats
from tabulate import tabulate as original_tabulate
import IPython


def is_needed_pairwise_combination(idx1, idx2, pairwise_mode):
    is_needed = True
    if 'not_identity' in pairwise_mode and idx1 == idx2:
        is_needed = False
    elif 'upper_triangle' in pairwise_mode and idx1 > idx2:
        is_needed = False
    elif 'lower_triangle' in pairwise_mode and idx1 < idx2:
        is_needed = False

    return is_needed


def tabulate_pairwise(data=None, config=None, **kwargs):
    """
    Plots a pairwise comparison between data from experiments based on a pairwise function d = f(exp_a, exp_b).

    :param data:
    :param config:
        tabulate: Parameters for the tabulate function that plots the table.
                  See https://pypi.org/project/tabulate/ for possible parameters.
                  Some important ones:

            tablefmt: Format of the table. (Default='html')
                        "plain", "simple", "github", "grid", "fancy_grid", "pipe", "orgtbl", "jira", "presto", "pretty",
                        "psql",  "rst", "mediawiki", "moinmoin", "youtrack", "html", "unsafehtml", "latex", "latex_raw",
                        "latex_booktabs", "latex_longtable", "textile", "tsv"

    :param kwargs:
    :return:
    """

    default_config = eu.AttrDict(

        pairwise_function = eu.misc.mannwhitneyu_pvalue,

        pairwise_mode = 'full',    # which pairs are compared? 'full', 'full_not_identity', 'upper_triangle', 'upper_triangle_not_identiy', 'lower_triangle', 'lower_triangle_not_identiy'

        tabulate=eu.AttrDict(
            tablefmt='html', #
            numalign='right',
        ),

        cell_format = '{}',

        top_left_cell_content = '',

        labels=[],  # holds all labels in a specific structure

    )
    config = eu.combine_dicts(kwargs, config, default_config)

    allowed_pairwise_modes = ['full', 'full_not_identity', 'upper_triangle', 'upper_triangle_not_identity', 'lower_triangle', 'lower_triangle_not_identity']
    if config.pairwise_mode not in allowed_pairwise_modes:
        raise ValueError('Unknown configuration {!r} for pairwise_mode! Allowed values: {}'.format(config.pairwise_mode, allowed_pairwise_modes))

    if data is None:
        data = np.array([])

    # format data in form [subplot_idx:list][trace_idx:list][elems_per_trace:numpy.ndarray]
    # subplot is a single table
    if isinstance(data, np.ndarray):
        data = [[data]]
    elif isinstance(data, list) and isinstance(data[0], np.ndarray):
        data = [data]
    elif not isinstance(data, list) and not isinstance(data[0], list) and not isinstance(data[0][0], np.ndarray):
        raise ValueError('Unsupported type of data!')

    # handle different input formats of labels
    if config.labels:
        # if only labels for mean-traces are given, then add an empty label for the sub figure
        if isinstance(config.labels, list) and not isinstance(config.labels[0], tuple):
            config.labels = [('', config.labels)]
        # if no labels are given for elements, then create an empty list for element labels
        for ds_idx in range(len(config.labels)):
            for trace_idx in range(len(config.labels[ds_idx][1])):
                if not isinstance(config.labels[ds_idx][1][trace_idx], tuple):
                    config.labels[ds_idx][1][trace_idx] = (config.labels[ds_idx][1][trace_idx], [])

    # identify the number of subplots
    n_subplots = len(data)

    if n_subplots > 1:
        raise NotImplementedError('Only supports 1 subplot!')

    # interate over subplots
    for subplot_idx, subplot_data in enumerate(data):

        trace_labels = []
        data_per_trace = []

        # collect the data and labels for each trace
        for trace_idx, cur_data in enumerate(subplot_data):

            # handle trace for mean values
            if config.labels:
                trace_label = config.labels[subplot_idx][1][trace_idx][0]
            else:
                trace_label = config.default_trace_label
                if len(config.trace_labels) > trace_idx:
                    trace_label = config.trace_labels[trace_idx]
            trace_label = eu.misc.replace_str_from_dict(str(trace_label), {'<trace_idx>': trace_idx})
            trace_labels.append(trace_label)

            data_points = np.array([])

            if np.ndim(cur_data) == 0:
                data_points = np.array([cur_data])
            elif np.ndim(cur_data) == 1:
                data_points = cur_data
            else:
                # collect data over elements
                for elem_idx, elem_data in enumerate(cur_data):  # data elements

                    # get element data which could be in matrix format or array format
                    if np.ndim(elem_data) == 0:
                        cur_elem_data = np.array([elem_data])
                    elif np.ndim(elem_data) == 1:
                        cur_elem_data = elem_data
                    elif np.ndim(elem_data) == 2:
                        if elem_data.shape[0] == 1:
                            cur_elem_data = elem_data[0, :]
                        elif elem_data.shape[1] == 1:
                            cur_elem_data = elem_data[1, 0]
                        else:
                            raise ValueError('Invalid data format!')
                    else:
                        raise ValueError('Invalid data format!')

                    data_points = np.hstack((data_points, cur_elem_data))

            data_per_trace.append(data_points)


        n_traces = len(data_per_trace)

        # compute the pairwise function of all needed combinations
        pairwise_data = np.full((n_traces,n_traces), np.nan)
        for first_trace_idx in range(n_traces):
            for second_trace_idx in range(n_traces):
                # decide if data has to be compared based on the config.pairwise_mode
                if is_needed_pairwise_combination(first_trace_idx, second_trace_idx, config.pairwise_mode):
                    pairwise_data[first_trace_idx, second_trace_idx] = eu.misc.call_function_from_config(
                        config.pairwise_function,
                        data_per_trace[first_trace_idx],
                        data_per_trace[second_trace_idx],
                    )

        # plot the results
        row_shift = 1
        col_shift = 1
        n_rows_and_cols = n_traces + 1
        # we leave some header out for these two cases
        if config.pairwise_mode == 'upper_triangle_not_identity':
            col_shift = 0
            n_rows_and_cols = n_traces
        elif config.pairwise_mode == 'lower_triangle_not_identity':
            row_shift = 0
            n_rows_and_cols = n_traces

        table_content = [[None] * (n_rows_and_cols) for _ in range(n_rows_and_cols)]

        table_content[0][0] = config.top_left_cell_content

        # set top and side header
        for trace_idx in range(n_traces):
            if config.pairwise_mode == 'upper_triangle_not_identity':
                # top header
                if trace_idx > 0:
                    table_content[0][trace_idx + col_shift] = trace_labels[trace_idx]
                # side header
                if trace_idx < n_traces - 1:
                    table_content[trace_idx + row_shift][0] = trace_labels[trace_idx]

            elif config.pairwise_mode == 'lower_triangle_not_identity':
                # top header
                if trace_idx < n_traces - 1:
                    table_content[0][trace_idx + col_shift] = trace_labels[trace_idx]
                # side header
                if trace_idx > 0:
                    table_content[trace_idx + row_shift][0] = trace_labels[trace_idx]

            else:
                # top header
                table_content[0][trace_idx + col_shift] = trace_labels[trace_idx]
                # side header
                table_content[trace_idx + row_shift][0] = trace_labels[trace_idx]

        # fill table
        for first_trace_idx in range(n_traces):
            for second_trace_idx in range(n_traces):
                if is_needed_pairwise_combination(first_trace_idx, second_trace_idx, config.pairwise_mode):

                    if isinstance(config.cell_format, str):
                        cell_data = config.cell_format.format(pairwise_data[first_trace_idx, second_trace_idx])
                    else:
                        cell_data = eu.misc.call_function_from_config(
                            config.cell_format,
                            pairwise_data[first_trace_idx, second_trace_idx])

                    table_content[first_trace_idx + row_shift][second_trace_idx + col_shift] = cell_data

        table = original_tabulate(
            table_content,
            headers='firstrow',
            **config.tabulate)

        return table


if __name__ == '__main__':

    # scalar values
    data = [[3, 5, 6]]
    data_labels = [('ds1', ['exp1', 'exp2', 'exp2'])]

    tabulate_pairwise(data, labels=data_labels)


##
## This file is part of the exputils package.
##
## Copyright: INRIA
## Year: 2022
## Contact: chris.reinke@inria.fr
##
## exputils is provided under GPL-3.0-or-later
##
import os
import stat
import re
import copy
import shutil
import exputils
from collections import OrderedDict

def generate_experiment_files(ods_filepath=None, directory=None, extra_files=None, extra_experiment_files=None, verbose=False):
    '''
    Generates experiment files and configurations based on entries in a ODS file (LibreOffice Spreadsheet).

    The ODS has to be in a specific form.
    Sheets define group of experiment for which an extra subfolder in the output directory will be generated.

    For an example of the ods format see the pytest files.

    :param ods_filepath: Path to the ODS file.
    :param directory: Directory where the experiments are generated.
    :param extra_files: Files that are mentioned in the ODS file but should be added to each experiment folder.
    '''

    if ods_filepath is None:
        ods_filepath = os.path.join('.', exputils.DEFAULT_ODS_CONFIGURATION_FILE)

    if directory is None:
        directory = os.path.join('.',exputils. DEFAULT_EXPERIMENTS_DIRECTORY)
    elif directory == '':
        directory = '.'

    if verbose:
        print('Load config from {!r} ...'.format(ods_filepath))

    config_data = load_configuration_data_from_ods(ods_filepath)

    # generate experiment files based on the loaded configurations
    if verbose:
        print('Generate experiments ...'.format(ods_filepath))

    generate_files_from_config(config_data, directory, extra_files=extra_files, extra_experiment_files=extra_experiment_files, verbose=verbose)


def load_configuration_data_from_ods(ods_filepath):
    '''

    :param ods_filepath:
    :return:
    '''

    config_data = []

    doc = exputils.io.ODSReader(ods_filepath, clonespannedcolumns=False)
    for sheet_name, sheet_data in doc.sheets.items():

        experiments_data = dict()

        experiments_data['directory'] = sheet_name
        experiments_data['experiments'] = OrderedDict()

        file_config = []

        parameters_start_idx = None
        parameters_end_idx = None

        file_borders = []

        variable_names = []

        #######################################################
        # first row: search for columns that describe parameters
        row_idx = 0

        for col_idx, col_data in enumerate(sheet_data[row_idx]):
            if col_data is not None:
                if col_data.lower() == 'parameters':
                    if parameters_start_idx is None:
                        parameters_start_idx = col_idx
                else:
                    parameters_end_idx = col_idx - 1
                    break

        #######################################################
        # second row: identify the template files
        row_idx = 1

        col_idx = parameters_start_idx
        stop = False

        previous_template_file_path = ''
        while not stop:

            if sheet_data[row_idx][col_idx] is not None and sheet_data[row_idx][col_idx] != previous_template_file_path:
                file_config.append(dict())
                file_config[-1]['template_file_path'] = sheet_data[row_idx][col_idx]

                previous_template_file_path = sheet_data[row_idx][col_idx]

                if file_borders:
                    file_borders[-1][1] = col_idx-1
                file_borders.append([col_idx, None])

            col_idx += 1
            if col_idx >= len(sheet_data[row_idx]) or (parameters_end_idx is not None and col_idx > parameters_end_idx):
                stop = True

        # in case the first line gives a definitive end, but not the second, then use the definitive end of the first line
        if parameters_end_idx is not None and file_borders[-1][1] is None:
            file_borders[-1][1] = parameters_end_idx

        #######################################################
        # third row: identify the file name templates
        row_idx = 2

        for file_idx in range(len(file_config)):
            file_config[file_idx]['file_name_template'] = sheet_data[row_idx][file_borders[file_idx][0]]

        #######################################################
        # fourth row: identify the variable names and if there are repetitions and a experiment folder
        row_idx = 3

        # find repetitions column
        repetitions_info_col_idx = None
        for col_idx in range(parameters_start_idx):
            if sheet_data[row_idx][col_idx] is not None and sheet_data[row_idx][col_idx].lower() == 'repetitions':
                repetitions_info_col_idx = col_idx
                break

        # find column with sources for repetitions
        repetition_source_file_locations_col_idx = None
        for col_idx in range(parameters_start_idx):
            if sheet_data[row_idx][col_idx] is not None and (sheet_data[row_idx][col_idx].lower() == 'files' or sheet_data[row_idx][col_idx].lower() == 'repetition files'):
                repetition_source_file_locations_col_idx = col_idx
                break
        # if no src directory for repetitions is given, then see if src/rep exists
        default_repetition_source_folder = None
        if not repetition_source_file_locations_col_idx:
            if os.path.isdir('./src/rep'):
                default_repetition_source_folder = './src/rep'

        # find column with sources for experiments
        experiment_source_file_locations_col_idx = None
        for col_idx in range(parameters_start_idx):
            if sheet_data[row_idx][col_idx] is not None and sheet_data[row_idx][col_idx].lower() == 'experiment files':
                experiment_source_file_locations_col_idx = col_idx
                break
        # if no src directory for repetitions is given, then see if src/rep exists
        default_experiment_source_folder = None
        if not experiment_source_file_locations_col_idx:
            if os.path.isdir('./src/exp'):
                default_experiment_source_folder = './src/exp'

        # find variable names
        for file_idx in range(len(file_config)):

            start_idx = file_borders[file_idx][0]
            end_idx = file_borders[file_idx][1]

            cur_variable_names = []

            col_idx = start_idx
            stop = False
            while not stop:
                cur_variable_names.append(sheet_data[row_idx][col_idx])

                col_idx += 1

                if col_idx < len(sheet_data[row_idx]) and sheet_data[row_idx][col_idx] is None:
                    file_borders[file_idx][1] = col_idx-1
                    stop = True

                elif col_idx >= len(sheet_data[row_idx]) or (end_idx is not None and col_idx > end_idx):
                    stop = True

            variable_names.append(cur_variable_names)

        #######################################################
        # remaining rows: experiments

        for row_idx in range(4, len(sheet_data)):

            experiment_id = sheet_data[row_idx][0]

            if experiment_id is not None:
                experiment_id = int(experiment_id)

                experiments_data['experiments'][experiment_id] = dict()

                experiments_data['experiments'][experiment_id]['files'] = copy.deepcopy(file_config)

                # repetitions info if it exists
                if repetitions_info_col_idx is not None and sheet_data[row_idx][repetitions_info_col_idx] is not None:
                    experiments_data['experiments'][experiment_id]['repetitions'] = int(sheet_data[row_idx][repetitions_info_col_idx])
                else:
                    experiments_data['experiments'][experiment_id]['repetitions'] = None

                if repetition_source_file_locations_col_idx is not None and sheet_data[row_idx][repetition_source_file_locations_col_idx] is not None:
                    experiments_data['experiments'][experiment_id]['repetition_source_file_locations'] = [i.strip() for i in sheet_data[row_idx][repetition_source_file_locations_col_idx].split(',')]
                elif default_repetition_source_folder:
                    experiments_data['experiments'][experiment_id]['repetition_source_file_locations'] = [default_repetition_source_folder]
                else:
                    experiments_data['experiments'][experiment_id]['repetition_source_file_locations'] = None

                if experiment_source_file_locations_col_idx is not None and sheet_data[row_idx][experiment_source_file_locations_col_idx] is not None:
                    experiments_data['experiments'][experiment_id]['experiment_source_file_locations'] = [i.strip() for i in sheet_data[row_idx][experiment_source_file_locations_col_idx].split(',')]
                elif default_experiment_source_folder:
                    experiments_data['experiments'][experiment_id]['experiment_source_file_locations'] = [default_experiment_source_folder]
                else:
                    experiments_data['experiments'][experiment_id]['experiment_source_file_locations'] = None

                for file_idx in range(len(file_config)):

                    col_idx = file_borders[file_idx][0]

                    experiments_data['experiments'][experiment_id]['files'][file_idx]['variables'] = dict()

                    for variable_name in variable_names[file_idx]:

                        if col_idx > len(sheet_data[row_idx]) - 1:
                            # if there is no content in the final cells, the array that holds the data is shorter
                            cur_cell_data = ''
                        else:
                            cur_cell_data = get_cell_data(sheet_data[row_idx][col_idx])

                        experiments_data['experiments'][experiment_id]['files'][file_idx]['variables'][variable_name] = cur_cell_data

                        col_idx +=1

        config_data.append(experiments_data)

    return config_data


def get_cell_data(data):

    if data is None:
        data = ''

    # replace strange characters that are not used for python strings
    data = data.replace('’', '\'')
    data = data.replace('`', '\'')

    return data


def generate_files_from_config(config_data, directory='.', extra_files=None, extra_experiment_files=None, verbose=False):
    '''

    Format of configuration data:

    config_data['directory']
    config_data['experiments']: dictionary with keys=experiment id, values=description:


    config_data['ids']: Ids of the experiments
    config_data['files']:  List with configuration for each template file
        template_file['template_file_path']: filepath of the template file
        template_file['file_name_template']: template for the filename
        template_file['variables']: Dictionary with key=variable name, value=variable value

    :param config_data:
    :return:
    '''

    if extra_files is None:
        extra_files = []
    elif not isinstance(extra_files, list):
        extra_files = [extra_files]

    if extra_experiment_files is None:
        extra_experiment_files = []
    elif not isinstance(extra_experiment_files, list):
        extra_experiment_files = [extra_experiment_files]

    for experiment_group_config in config_data:

        # only create group folder if more than one sheet or the sheet name is not empty or 'Sheet1'
        if len(config_data) == 1 and experiment_group_config['directory'] in ['Sheet1', '']:
            group_directory = directory
        else:
            group_directory = os.path.join(directory, experiment_group_config['directory'])

        # create the folder if not exists
        if not os.path.isdir(group_directory):
            os.makedirs(group_directory)

        # generate the experiment folders and files
        for experiment_id, experiment_config in experiment_group_config['experiments'].items():

            experiment_directory = os.path.join(group_directory, exputils.EXPERIMENT_DIRECTORY_TEMPLATE.format(experiment_id))

            # create folders for the repetitions if necessary:
            if experiment_config['repetitions'] is None:
                num_of_repetitions = 1
            else:
                num_of_repetitions = experiment_config['repetitions']

            for repetition_id in range(num_of_repetitions):

                # only create repetition folders if repetitions are specified
                if experiment_config['repetitions'] is None:
                    experiment_files_directory = experiment_directory
                else:
                    experiment_files_directory = os.path.join(experiment_directory, exputils.REPETITION_DIRECTORY_TEMPLATE.format(repetition_id))

                # create folder if not exists
                if not os.path.isdir(experiment_files_directory):
                    os.makedirs(experiment_files_directory)

                # generate the files for the experiment, or the repetition if they are defined
                if experiment_config['repetition_source_file_locations'] is None:
                    source_files = extra_files
                else:
                    source_files = experiment_config['repetition_source_file_locations'] + extra_files

                generate_source_files(source_files, experiment_files_directory, experiment_config, experiment_id, repetition_id)

            # if there are experiment - repetitions defined, then generate the files for the experiment folder
            if experiment_config['experiment_source_file_locations'] is None:
                source_files = extra_experiment_files
            else:
                source_files = experiment_config['experiment_source_file_locations'] + extra_experiment_files

            if source_files:
                generate_source_files(source_files, experiment_directory, experiment_config, experiment_id)


def generate_source_files(source_files, experiment_files_directory, experiment_config, experiment_id, repetition_id=None):

    # create the source files that are given by templates
    for file_config in experiment_config['files']:

        # check if the given file exists, otherwise it might be in one of the given directories under the source-files property of the experiment
        template_file_path = None
        if os.path.isfile(file_config['template_file_path']):
            template_file_path = file_config['template_file_path']
        else:
            for src in source_files:
                if os.path.isdir(src):
                    if os.path.isfile(os.path.join(src, file_config['template_file_path'])):
                        template_file_path = os.path.join(src, file_config['template_file_path'])
                        break

        if template_file_path is not None:

            # get file permissions
            permissions = os.stat(template_file_path)[stat.ST_MODE]

            # Read in the template file
            with open(template_file_path, 'r') as file:
                file_content = file.read()

            # Replace the variables
            file_content = file_content.replace('<experiment_id>', str(experiment_id))

            if repetition_id is not None:
                file_content = file_content.replace('<repetition_id>', str(repetition_id))

            for variable_name, variable_value in file_config['variables'].items():
                file_content = re.sub('<{}>'.format(variable_name),
                                      variable_value,
                                      file_content,
                                      flags=re.IGNORECASE)

            # Write the final output file
            file_path = os.path.join(experiment_files_directory, file_config['file_name_template'].format(experiment_id))
            with open(file_path, 'w') as file:
                file.write(file_content)
            os.chmod(file_path, permissions)

    # copy all other sources, but not the templates if they are in one of the source directories
    template_files = [file_config['template_file_path'] for file_config in experiment_config['files']]

    for src in source_files:
        copy_experiment_files(src, experiment_files_directory, template_files)


def copy_experiment_files(src, dst, template_files):

    if os.path.isdir(src):
        # if directory, then copy the content

        for item in os.listdir(src):
            s = os.path.join(src, item)

            # if subdirectory, then delete any existing directory and make a new directory

            if os.path.isdir(s):

                d = os.path.join(dst, item)

                if os.path.isdir(d):
                    shutil.rmtree(d, ignore_errors=True)

                os.mkdir(d)

            else:
                d = dst

            copy_experiment_files(s, d, template_files)

    else:
        # if file, then copy it directly

        # do not copy the template files, because they were already processed
        if os.path.basename(src) not in [os.path.basename(f) for f in template_files]:
            shutil.copy2(src, dst)






##
## This file is part of the exputils package.
##
## Copyright: INRIA
## Year: 2022
## Contact: chris.reinke@inria.fr
##
## exputils is provided under GPL-3.0-or-later
##
import numpy as np


def get_ordered_experiment_ids_from_descriptions(experiment_descriptions):
    """
    Returns a sorted list of experiment ids according to the order in the experiment descriptions.
    """
    if not experiment_descriptions:
        return []

    experiment_descriptions_values = list(experiment_descriptions.values())
    order = [descr['order'] for descr in experiment_descriptions_values]
    sorted_experiment_ids = [experiment_descriptions_values[idx]['id'] for idx in np.argsort(order)]
    return sorted_experiment_ids


##
## This file is part of the exputils package.
##
## Copyright: INRIA
## Year: 2022
## Contact: chris.reinke@inria.fr
##
## exputils is provided under GPL-3.0-or-later
##
# Taken from https://pypi.org/project/ipynbname/
# TODO: add them to licence

from notebook import notebookapp
import urllib, json, os, ipykernel, ntpath

FILE_ERROR = "Can't identify the notebook {}."
CONN_ERROR = "Unable to access server;\n \
            + ipynbname requires either no security or token based security."

def _get_kernel_id():
    """ Returns the kernel ID of the ipykernel.
    """
    connection_file = os.path.basename(ipykernel.get_connection_file())
    kernel_id = connection_file.split('-', 1)[1].split('.')[0]
    return kernel_id


def _get_sessions(srv):
    """ Given a server, returns sessions, or HTTPError if access is denied.
        NOTE: Works only when either there is no security or there is token
        based security. An HTTPError is raised if unable to connect to a
        server.
    """
    try:
        qry_str = ""
        token = srv['token']
        if token:
            qry_str = f"?token={token}"
        url = f"{srv['url']}api/sessions{qry_str}"
        req = urllib.request.urlopen(url)
        return json.load(req)
    except:
        raise urllib.error.HTTPError(CONN_ERROR)


def _get_nb_path(sess, kernel_id):
    """ Given a session and kernel ID, returns the notebook path for the
        session, or None if there is no notebook for the session.
    """
    if sess['kernel']['id'] == kernel_id:
        return sess['notebook']['path']
    return None


def get_notebook_name():
    """ Returns the short name of the notebook w/o the .ipynb extension,
        or raises a FileNotFoundError exception if it cannot be determined.
    """
    kernel_id = _get_kernel_id()
    for srv in notebookapp.list_running_servers():
        try:
            sessions = _get_sessions(srv)
            for sess in sessions:
                nb_path = _get_nb_path(sess, kernel_id)
                if nb_path:
                    return ntpath.basename(nb_path).replace('.ipynb', '')
        except:
            pass  # There may be stale entries in the runtime directory
    raise FileNotFoundError(FILE_ERROR.format('name'))


def get_notebook_path():
    """ Returns the absolute path of the notebook,
        or raises a FileNotFoundError exception if it cannot be determined.
    """
    kernel_id = _get_kernel_id()
    for srv in notebookapp.list_running_servers():
        try:
            sessions = _get_sessions(srv)
            for sess in sessions:
                nb_path = _get_nb_path(sess, kernel_id)
                if nb_path:
                    return os.path.join(srv['notebook_dir'], nb_path)
        except:
            pass  # There may be stale entries in the runtime directory
    raise FileNotFoundError(FILE_ERROR.format('path'))


##
## This file is part of the exputils package.
##
## Copyright: INRIA
## Year: 2022
## Contact: chris.reinke@inria.fr
##
## exputils is provided under GPL-3.0-or-later
##
import exputils as eu
import ipywidgets
from exputils.gui.jupyter.experiment_data_selection_widget import ExperimentDataSelectionWidget
import IPython


DEFAULT_PLOTLY_MEANSTD_SCATTER_CONFIG = """layout = dict(
    xaxis = dict(
        title = '', 
        range = [None, None]),
    yaxis = dict(
        title = '', 
        range = [None, None]),
),
moving_average = dict(
    n = 1
),
data_filter = dict(
    every_nth_step = dict(
        step = 1,
        include_final_step = False
    ),
),        
default_group_label = 'rep <group_idx>'"""

DEFAULT_PLOTLY_BOX_CONFIG = """layout = dict(
    xaxis = dict(
        title = '', 
        range = [None, None] 
        ),
    yaxis = dict(
        title = '', 
        range = [None, None] 
        )
    )"""

DEFAULT_PLOTLY_MEANSTD_BAR_CONFIG = """layout = dict(
    xaxis = dict(
        title = '', 
        range = [None, None] 
        ),
    yaxis = dict(
        title = '', 
        range = [None, None] 
        )
    ),
    default_group_label = 'rep <group_idx>'"""

DEFAULT_TABULATE_MEANSTD_CONFIG = """primary_content_function = np.nanmean,
secondary_content_function = np.nanstd,
flip_rows_and_cols = False,
tabulate=eu.AttrDict(
            tablefmt='html', #
            numalign='right',
        ),
cell_format = '{:.3f} ({:.3f})',
top_left_cell_content = ''"""

DEFAULT_TABULATE_PAIRWISE_CONFIG = """pairwise_function = eu.misc.mannwhitneyu_pvalue,
pairwise_mode = 'upper_triangle', # which pairs are compared? 'full', 'full_not_identity', 'upper_triangle', 'upper_triangle_not_identity', 'lower_triangle', 'lower_triangle_not_identity'
tabulate=eu.AttrDict(
            tablefmt='html', # 'html' or 'latex'
            numalign='right',
        ),
cell_format = '{}',
top_left_cell_content = 'p-value',
labels=[]"""

CODE_TEMPLATE_MULTILINE = """# Plotting of <datasources> 
import exputils as eu
<import_statements>

plot_config = <plot_function_config>

selection_widget = eu.gui.jupyter.ExperimentDataPlotSelectionWidget(
    experiment_data_loader,
    datasources=<datasources>,
    experiment_ids='all',
    repetition_ids='all',
    output_format=<output_format>,
    data_filter=<data_filter>,
    plot_function=<plot_function>,
    plot_function_config=plot_config,
    state_backup_name=<state_backup_name>,
    state_backup_variable_filter=['experiment_ids', 'repetition_ids'],  # only save these variables as backup
    is_datasources_selection=False,
    is_output_format_selection=False,
    is_data_filter_selection=False,
    is_plot_function_selection=False,
    is_plot_function_config_editor=False,
    is_code_producer=False) 
display(selection_widget)
selection_widget.plot_data()"""

CODE_TEMPLATE_SINGLELINE = """# Plotting of <datasources> 
import exputils as eu
<import_statements>

plot_config = <plot_function_config>

selection_widget = eu.gui.jupyter.ExperimentDataPlotSelectionWidget(experiment_data_loader, datasources=<datasources>, experiment_ids='all', repetition_ids='all', output_format=<output_format>, data_filter=<data_filter>, plot_function=<plot_function>, plot_function_config=plot_config, state_backup_name=<state_backup_name>, state_backup_variable_filter=['experiment_ids', 'repetition_ids'], is_datasources_selection=False, is_output_format_selection=False, is_data_filter_selection=False, is_plot_function_selection=False, is_plot_function_config_editor=False, is_code_producer=False)
display(selection_widget)
selection_widget.plot_data()"""

def config_obj_to_dict(config_obj):
    '''
    Transforms a potential configuration (either dict or string) to configuration dictionary.
    If it is a string, it tries to execute the <string> as python code with 'dict(<string>)'.
    If this fails, then the returned object the Exception.
    '''
    if config_obj is None:
        config = None
    elif isinstance(config_obj, dict):
        config = config_obj
    else:
        try:
            _locals = {}
            exec('import numpy as np;import exputils as eu;tmp_dict = dict({})'.format(config_obj), {}, _locals)
            config = _locals['tmp_dict']
        except Exception as err:
            config = err
    return config


class ExperimentDataPlotSelectionWidget(ExperimentDataSelectionWidget):

    @staticmethod
    def default_config():
        dc = ExperimentDataSelectionWidget.default_config()

        # do not show the get data button, this widget will create its own plotting button
        dc.is_get_experiment_data_button = False

        # dictionary with possible plotting function
        dc.plot_functions = {'plotly_meanstd_scatter': eu.gui.jupyter.plotly_meanstd_scatter,
                             'plotly_box': eu.gui.jupyter.plotly_box,
                             'plotly_meanstd_bar': eu.gui.jupyter.plotly_meanstd_bar,
                             'tabulate_meanstd': eu.gui.jupyter.tabulate_meanstd,
                             'tabulate_pairwise': eu.gui.jupyter.tabulate_pairwise,

        }

        # dictionary with plot_function_configs for each
        dc.plot_function_configs = {'plotly_meanstd_scatter': DEFAULT_PLOTLY_MEANSTD_SCATTER_CONFIG,
                                    'plotly_box': DEFAULT_PLOTLY_BOX_CONFIG,
                                    'plotly_meanstd_bar': DEFAULT_PLOTLY_MEANSTD_BAR_CONFIG,
                                    'tabulate_meanstd': DEFAULT_TABULATE_MEANSTD_CONFIG,
                                    'tabulate_pairwise': DEFAULT_TABULATE_PAIRWISE_CONFIG,
        }

        dc.is_plot_function_selection = True
        dc.plot_function = list(dc.plot_functions.keys())[0]
        dc.plot_function_selection = eu.AttrDict(
            hbox=eu.AttrDict(  # ipywidgets.HBox parameters
                layout=eu.AttrDict(
                    width='100%')),
            label=eu.AttrDict(  # ipywidgets.Label parameters
                value='Plot Function:',
                layout=eu.AttrDict(min_width='100px')),
            dropdown=eu.AttrDict(  # ipywidgets.Text parameters
                description='',
                layout=eu.AttrDict(width='100%')))

        dc.is_plot_function_config_editor = True
        dc.plot_function_config = None
        dc.plot_function_config_editor = eu.AttrDict(
            title = 'Plot Configuration',
            accordion=eu.AttrDict(  # ipywidgets.HBox parameters
                selected_index=None,  # collapse accordion at init
                layout=eu.AttrDict(
                    width='100%')),
            textarea=eu.AttrDict(  # ipywidgets.Label parameters
                placeholder='Provide the configuration of the plot function in form of a python dictionary ...',
                layout=eu.AttrDict(
                    min_width='100%'))) # TODO: start with a larger height at the beginning

        dc.is_plot_button = True
        dc.plot_button = eu.AttrDict(
            description='Plot Data',
            tooltip='Plots the data according to the selection.',
            layout=eu.AttrDict(width='100%', height='95%'))

        dc.code_producer.code_templates = [dict(name='Multi Line', code_template=CODE_TEMPLATE_MULTILINE),
                                           dict(name='Single Line', code_template=CODE_TEMPLATE_SINGLELINE)]

        dc.figure_output = eu.AttrDict()  # ipywidgets.Output parameters

        return dc


    def __init__(self, experiment_data, experiment_descriptions=None, config=None, **kwargs):

        super().__init__(experiment_data,
                         experiment_descriptions=experiment_descriptions,
                         config=config,
                         **kwargs)

        # add a figure output below the selection
        self.figure_output = None

        #########################
        # handle config input if not a state backup was loaded
        if not hasattr(self, '_plot_function_key') and not hasattr(self, '_plot_function') and not hasattr(self, '_plot_function_configs'):

            # use first plot function in plot_fuctions dictionary if non is provided
            if self.config.plot_function == '' or self.config.plot_function is None:
                plot_function = list(self.config.plot_functions.keys())[0]
            else:
                plot_function = self.config.plot_function

            if isinstance(plot_function, str):
                # config.plot_function is a key for the config.plot_functions dictionary

                if self.config.plot_function not in self.config.plot_functions:
                    raise ValueError('If config.plot_function is a string then it must be a key for the config.plot_functions dictionary!')

                self._plot_function_key = plot_function
                self._plot_function = self.config.plot_functions[self._plot_function_key]

            elif callable(plot_function):
                # if the config.plotfunction is a function handle
                self._plot_function_key = 'NONE'
                self._plot_function = plot_function

            else:
                raise TypeError('Unsupported type for config.plot_fuction! Only strings or function handles are valid.')

            # handle self._plot_function_configs
            self._plot_function_configs = self.config.plot_function_configs.copy()
            plot_function_config = self.config.plot_function_config
            if plot_function_config == '' or plot_function_config is None:
                # if no config.plot_function_config is given, then use the one from the config.plot_function_configs dictionary
                # create a new entry in _plot_function_configs if non exists for this plot_function_key
                if self._plot_function_key not in self._plot_function_configs:
                    self._plot_function_configs[self._plot_function_key] = None
            else:
                # if a config is given, then override the _plot_functions_config with it
                self._plot_function_configs[self._plot_function_key] = config_obj_to_dict(plot_function_config)

            # if the config is invalid and the user can not change it anymore, then raise an exception
            if not self.config.is_plot_function_config_editor \
                    and isinstance(self._plot_function_configs[self._plot_function_key], BaseException):
                raise self._plot_function_config[self._plot_function_key]

            if self._plot_function_key == 'NONE' and self.config.is_plot_function_selection:
                raise ValueError('If plot_function_selection is active, then the config.plot_function must be a key for the config.plot_functions dictionary!')

        ########################
        # add gui components
        selection_children = []

        # only allow selection of a plot function if the defined intial plot_function is not a function handle
        if self.config.is_plot_function_selection and callable(self.config.plot_function):
            self.config.is_plot_function_selection = False

        # selection of plot function
        if self.config.is_plot_function_selection:

            self.plot_function_selection_label_widget = ipywidgets.Label(**self.config.plot_function_selection.label)

            self.plot_function_selection_dropdown_widget = ipywidgets.Dropdown(
                options=list(self.config.plot_functions.keys()),
                value=self._plot_function_key,
                **self.config.plot_function_selection.dropdown)

            self.plot_function_selection_hbox_widget = ipywidgets.HBox(
                [self.plot_function_selection_label_widget, self.plot_function_selection_dropdown_widget],
                **self.config.plot_function_selection.hbox)

            selection_children.append(self.plot_function_selection_hbox_widget)

            # register event to know when a new plot function was selected
            self.plot_function_selection_dropdown_widget.observe(
                self._on_plot_function_selection_dropdown_widget_value_change,
                names='value')

        if self.config.is_plot_function_config_editor:

            self.plot_function_config_editor_textarea = ipywidgets.Textarea(
                value=self._plot_function_configs[self._plot_function_key],
                **self.config.plot_function_config_editor.textarea)

            self.plot_function_config_editor_accordion = ipywidgets.Accordion(
                children=[self.plot_function_config_editor_textarea],
                **self.config.plot_function_config_editor.accordion)
            self.plot_function_config_editor_accordion.set_title(0, self.config.plot_function_config_editor.title)

            selection_children.append(self.plot_function_config_editor_accordion)

            # register event to know when a new config was edited
            self.plot_function_config_editor_textarea.observe(
                self._on_plot_function_config_editor_value_change,
                names='value')

        # add selection items befor the box for the buttons
        eu.gui.jupyter.add_children_to_widget(self, selection_children, idx=-1)

        # add plotting button
        if self.config.is_plot_button:
            self.plot_button = ipywidgets.Button(**self.config.plot_button)
            self.plot_button.on_click(self._plot_button_on_click_handler)

            # append button to start of button box
            eu.gui.jupyter.add_children_to_widget(self.activity_hbox, self.plot_button, idx=0)


    def _on_plot_function_selection_dropdown_widget_value_change(self, event_descr):
        self._plot_function_key = event_descr['new']
        self._plot_function = self.config.plot_functions[self._plot_function_key]
        self.plot_function_config = self._plot_function_configs.get(self._plot_function_key, None)


    def _on_plot_function_config_editor_value_change(self, event_descr):
        self._plot_function_configs[self._plot_function_key] = event_descr['new']


    @property
    def plot_function(self):
        return self._plot_function


    @plot_function.setter
    def plot_function(self, plot_function):

        if isinstance(plot_function, str):
            # config.plot_function is a key for the config.plot_functions dictionary

            if plot_function not in self.config.plot_functions:
                raise ValueError(
                    'If plot_function is a string then it must be a key for the config.plot_functions dictionary!')

            self._plot_function_key = plot_function
            self._plot_function = self.config.plot_functions[self._plot_function_key]

            # also the config.plot_function_configs dict will be used
            self.plot_function_config = self._plot_function_configs[self._plot_function_key]

        elif callable(plot_function):
            # if the config.plotfunction is a function handle

            # try to find function in configured functions
            plot_function_key = 'NONE'
            for func_str, func in self.config.plot_functions.items():
                if func == plot_function:
                    plot_function_key = func_str
                    break

            self._plot_function_key = plot_function_key
            self._plot_function = plot_function
            if self._plot_function_key not in self._plot_function_configs:
                self._plot_function_configs[self._plot_function_key] = None

    @property
    def plot_function_config(self):
        return config_obj_to_dict(self._plot_function_configs[self._plot_function_key])


    @plot_function_config.setter
    def plot_function_config(self, plot_function_config):

        if self.config.is_plot_function_config_editor:

            if not isinstance(plot_function_config, str):
                raise ValueError('If the plot_function_config_editor is activated, then the given plot_function_config must be a string!')

            # was the gui element already created?
            if hasattr(self, 'plot_function_config_editor_textarea'):
                self.plot_function_config_editor_textarea.value = plot_function_config

        self._plot_function_configs[self._plot_function_key] = plot_function_config


    def _plot_button_on_click_handler(self, _):
        self.plot_data()


    @property
    def selection(self):
        '''AttrDict (dict) with the selected experiment data options.'''
        selection = super().selection
        selection.plot_function = self.plot_function
        selection.plot_function_config = self.plot_function_config
        return selection


    @selection.setter
    def selection(self, selection):
        super(self.__class__, self.__class__).selection.fset(self, selection)
        if 'plot_function' in selection: self.plot_function = selection.plot_function
        if 'plot_function_config' in selection: self.plot_function_config = selection.plot_function_config


    def get_widget_state(self):
        state = super().get_widget_state()

        if self._plot_function_key == 'NONE':
            state.plot_function = self.plot_function
        else:
            state.plot_function = self._plot_function_key
        state._plot_function_configs = self._plot_function_configs
        return state


    def set_widget_state(self, state):

        if '_plot_function_configs' in state:
            self._plot_function_configs = state._plot_function_configs  # set all gui configs

        if 'plot_function' in state:
            self.plot_function = state.plot_function

        if '_plot_function_configs' in state:
            self.plot_function_config = self._plot_function_configs[self._plot_function_key]  # update the view

        return super().set_widget_state(state)


    def plot_data(self):

        # display the figure
        # must be created here so that the display is created below the selection gui
        if self.figure_output is None:
            self.figure_output = ipywidgets.Output(**self.config.figure_output)
            IPython.display.display(self.figure_output)
        else:
            self.figure_output.clear_output(wait=True)

        with self.figure_output:
            print('Plotting ...')

            # load experimental data before plotting
            self.select_experiment_data()

            plot_config = self.plot_function_config
            if isinstance(plot_config, BaseException):
                raise plot_config
            else:
                display_obiect = self.plot_function(
                    self.selected_data,
                    labels=self.selected_data_labels,
                    config=plot_config)

                self.figure_output.clear_output(wait=True)
                IPython.display.display(display_obiect)


    def get_code_producer_variables(self):

        variables = super().get_code_producer_variables()

        plot_function_handle = self.plot_function
        variables.plot_function = plot_function_handle.__name__

        if plot_function_handle.__module__ == '__main__' or plot_function_handle.__module__ == 'builtins':
            # no extra imports
            variables.import_statements = '\n'
        else:
            variables.import_statements = 'from {} import {}'.format(plot_function_handle.__module__, variables.plot_function)

        # plot configuration as a dictionary
        if self.config.is_plot_function_config_editor:
            plot_function_config_str = 'eu.AttrDict(\n{})'.format(self.plot_function_config_editor_textarea.value)
        else:
            plot_function_config_dict = self.plot_function_config
            plot_function_config_str = 'eu.' + str(eu.AttrDict(plot_function_config_dict))

        variables.plot_function_config = plot_function_config_str

        return variables


##
## This file is part of the exputils package.
##
## Copyright: INRIA
## Year: 2022
## Contact: chris.reinke@inria.fr
##
## exputils is provided under GPL-3.0-or-later
##
import exputils as eu
import dill
import os
from glob import glob

DILL_FILE_EXTENSION = 'dill'

def save_dill(obj, file_path):

    if not file_path.endswith('.' + DILL_FILE_EXTENSION):
        file_path += '.' + DILL_FILE_EXTENSION

    eu.io.makedirs_for_file(file_path)
    with open(file_path, 'wb') as fh:
        dill.dump(obj, fh)


def load_dill(file_path):

    if not os.path.exists(file_path):
        if not file_path.endswith('.' + DILL_FILE_EXTENSION):
            file_path += '.' + DILL_FILE_EXTENSION

    with open(file_path, 'rb') as fh:
        obj = dill.load(fh)
    return obj


def load_dill_files(directory):
    '''Loads data from all dill files in a given directory.'''

    if not os.path.isdir(directory):
        raise FileNotFoundError('Directory {!r} does not exist!'.format(directory))

    data_dict = eu.AttrDict()

    for file in glob(os.path.join(directory, '*.' + DILL_FILE_EXTENSION)):
        data_name = os.path.splitext(os.path.basename(file))[0]
        data = load_dill(file)
        data_dict[data_name] = data

    return data_dict


##
## This file is part of the exputils package.
##
## Copyright: INRIA
## Year: 2022
## Contact: chris.reinke@inria.fr
##
## exputils is provided under GPL-3.0-or-later
##
import glob
import os
import subprocess
import time
import exputils
import numpy as np
from datetime import datetime
import fasteners

STATUS_FILE_EXTENSION = '.status'


def start_slurm_experiments(directory=None, start_scripts='*.slurm', is_parallel=True, verbose=False, post_start_wait_time=0):

    return start_experiments(
        directory=directory,
        start_scripts=start_scripts,
        start_command='sbatch {}',
        parallel=is_parallel,
        verbose=verbose,
        post_start_wait_time=post_start_wait_time,
        write_status_files_automatically=False
    )


def start_torque_experiments(directory=None, start_scripts='*.torque', is_parallel=True, verbose=False, post_start_wait_time=0):

    return start_experiments(
        directory=directory,
        start_scripts=start_scripts,
        start_command='qsub {}',
        parallel=is_parallel,
        verbose=verbose,
        post_start_wait_time=post_start_wait_time,
        write_status_files_automatically=False
    )


def start_experiments(directory=None, start_scripts='*.sh', start_command='{}', parallel=True, is_chdir=True, verbose=False, post_start_wait_time=0,
                      write_status_files_automatically=True):
    """
    Starts experiments and repetitions in a parallel.

    :param directory: Directory in which the start scripts are searched.
    :param start_scripts: Filename of the start script file. Can include * to search for scripts.
    :param start_command:
    :param parallel: True if processes are started and executed in parallel. False if they are executed one after the other.
                     A integer number defines how many processes can run in parallel.
    :param is_chdir: Before starting a script, should the main process change to its working directory. (Default: True)
    :param verbose:
    :param post_start_wait_time:
    :param write_status_files_automatically: Should the status file for the started scripts be written automatically by exputils or by the started
                                             process itself. (Default: True)
    :return:
    """

    # TODO: remove the is_rerun argument, as it is not used anymore. This makes also changes in the commands necessary!

    if directory is None:
        directory = os.path.join('.', exputils.DEFAULT_EXPERIMENTS_DIRECTORY)

    # handle number of parallel processes
    if isinstance(parallel, bool):
        if parallel:
            n_parallel = np.inf
        else:
            n_parallel = 1
    elif isinstance(parallel, int):
        if parallel <= 0:
            raise ValueError('Number of parallel processes must be larger 0!')
        else:
            n_parallel = parallel
    else:
        raise ValueError('Argument \'parallel\' must be either a bool or an integer number!')

    if is_chdir:
        cwd = os.getcwd()

    # get all scripts
    all_scripts = get_scripts(directory=directory, start_scripts=start_scripts)

    ignored_scripts = []
    todo_scripts = []
    # check their initial status and write one for the scripts that will be started
    for script in all_scripts:

        # lock processing of the script, so that no other running experimentstarter is updating its status in parallel
        with get_script_lock(script):

            status = get_script_status(script)

            if status is None:
                if write_status_files_automatically:
                    update_script_status(script, 'todo')
                todo_scripts.append(script)

            elif status.lower() != 'finished':
                todo_scripts.append(script)

            else:
                ignored_scripts.append((script, status))

    # start all in parallel if wanted
    if n_parallel == np.inf:
        n_parallel = len(todo_scripts)

    # started process and their corresponding scripts
    started_processes = []
    started_scripts = []
    finished_processes_idxs = []

    next_todo_script_idx = 0
    n_active_processes = 0

    # run as long as there is an active process or we did not finish all processes yet
    while n_active_processes > 0 or next_todo_script_idx < len(todo_scripts):

        # start as many processes as parallel processes are allowed
        for i in range(n_parallel - n_active_processes):

            # stop starting processes when all scripts are started
            if next_todo_script_idx < len(todo_scripts):

                script = todo_scripts[next_todo_script_idx]
                next_todo_script_idx += 1

                # lock processing of the script, so that no other running experimentstarter is starting it in parallel
                with get_script_lock(script):

                    # check the script status, only start if needed
                    status = get_script_status(script)
                    if is_to_start_status(status):

                        if write_status_files_automatically:
                            update_script_status(script, 'running')

                        # start
                        script_directory = os.path.dirname(script)
                        script_path_in_its_working_directory = os.path.join('.', os.path.basename(script))

                        print('{} start {!r} (previous status: {}) ...'.format(datetime.now().strftime("%Y/%m/%d %H:%M:%S"), script, status))

                        process_environ = {
                            **os.environ,
                            "EU_STATUS_FILE": script_path_in_its_working_directory + STATUS_FILE_EXTENSION,
                        }

                        if is_chdir:
                            os.chdir(script_directory)
                            process = subprocess.Popen(start_command.format(script_path_in_its_working_directory).split(), env=process_environ)
                            os.chdir(cwd)
                        else:
                            process = subprocess.Popen(start_command.format(script).split(), cwd=script_directory, env=process_environ)

                        started_processes.append(process)
                        started_scripts.append(script)

                        if post_start_wait_time > 0:
                            time.sleep(post_start_wait_time)

                    else:
                        # do not start
                        ignored_scripts.append((script, status))

        # check the activity of the started processes
        n_active_processes = 0
        for p_idx, process in enumerate(started_processes):

            if p_idx not in finished_processes_idxs:

                if process.poll() is None:
                    n_active_processes += 1
                else:
                    finished_processes_idxs.append(p_idx)
                    if process.returncode == 0:
                        status = 'finished'
                    else:
                        status = 'error'

                    if write_status_files_automatically:
                        update_script_status(started_scripts[p_idx], status)

                    print('{} finished {!r} (status: {})'.format(datetime.now().strftime("%Y/%m/%d %H:%M:%S"), started_scripts[p_idx], status))

        if n_active_processes > 0:
            time.sleep(0.5) # sleep half a second before checking again

    if verbose:
        if ignored_scripts:
            print('Ignored scripts:')
            for (script_path, status) in ignored_scripts:
                print('\t- {!r} (status: {})'.format(script_path, status))


def is_to_start_status(status):
    """Returns true if the given status means that the script should be started, otherwise false."""
    return status is None or status.lower().startswith('todo') or status.lower().startswith('none') or status.lower().startswith('error') or status.lower().startswith('unfinished')


def get_script_lock(script):
    """Create a lock for the given script that can be used to have exclusive access to write its status."""
    return fasteners.InterProcessLock(script + '.lock')


def update_script_status(script, status):
    """
    Updates the status for the given script.

     :param script: Path to the script.
     :param status: New status.
    """

    status_file_path = script + STATUS_FILE_EXTENSION

    time_str = datetime.now().strftime("%Y/%m/%d %H:%M:%S")

    with open(status_file_path, 'a+') as file:
        file.write( time_str + "\n" + status + "\n")


def get_scripts(directory=None, start_scripts='*.sh'):
    """
    Idenitfies all scripts.

     :param directory: Directory in which the start scripts are searched.
     :param start_scripts: Filename of the start script file. Can include * to search for scripts.

     :return: List of scripts (pathes).
    """

    if directory is None:
        directory = os.path.join('.', exputils.DEFAULT_EXPERIMENTS_DIRECTORY)

    # find all start scripts
    scripts = glob.iglob(os.path.join(directory, '**', start_scripts), recursive=True)
    scripts = list(scripts)
    scripts.sort()

    return scripts


def get_script_status(script_file):
    """
    Returns the status of the given script file.

    :param script_file: Path to the script file.

    :return: Status as a string. ('error', 'running', 'finished'). None if not status exists.
    """
    status = None

    status_file_path = script_file + STATUS_FILE_EXTENSION

    if os.path.isfile(status_file_path):
        # read status
        with open(status_file_path, 'r') as f:
            lines = f.read().splitlines()
            if len(lines) > 0:
                status = lines[-1]

    return status


def get_number_of_scripts_to_execute(directory=None, start_scripts='*.sh'):
    """
    Returns the number scripts that have to be executed

    :param directory: Directory in which the start scripts are searched.
    :param start_scripts: Filename of the start script file. Can include * to search for scripts.

    :return: Number of scripts that have to be executed (int).
    """

    scripts = get_scripts(directory=directory, start_scripts=start_scripts)

    n = 0
    for script in scripts:
        status = get_script_status(script)
        if is_to_start_status(status):
            n += 1

    return n


def get_number_of_scripts(directory=None, start_scripts='*.sh'):
    """
    Returns the number scripts of all scripts regardless of their run status.

    :param directory: Directory in which the start scripts are searched.
    :param start_scripts: Filename of the start script file. Can include * to search for scripts.

    :return: Number of scripts (int).
    """

    scripts = get_scripts(directory=directory, start_scripts=start_scripts)
    return len(scripts)







##
## This file is part of the exputils package.
##
## Copyright: INRIA
## Year: 2022
## Contact: chris.reinke@inria.fr
##
## exputils is provided under GPL-3.0-or-later
##
import numpy as np
import plotly
import plotly.subplots
import exputils as eu

# TODO: Feature - allow to first unselect certain experiments, and then switch to their elements, to just see the selected experiments
#       https://webappl.blogspot.com/2020/05/plotly-eventregister.html, see plotly_restyle event
#       I believe I need to create a Figure object for this purpose
# TODO: Feature - custom x values

def plotly_meanstd_scatter(data=None, config=None, **kwargs):
    """
    Plots a plotly scatter (Scattergl or Scatter) plot.

    :param data: Data to plot. Should be in the following forms:
        - [subplot_idx:list][trace_idx:list][elems_per_trace:numpy.ndarray]
        - [trace_idx:list][elems_per_trace:numpy.ndarray]
        - elems_per_trace:numpy.ndarray

    :param config: Dictionary with configuration of plot.
        moving_average: Display the moving average over the steps per trace-element.
            n: number of steps over which the average should be computed.

        data_filter:
            every_nth_step: Either a integer with the number of steps or a dictionary.
                            In case of a dictionary:
                step: Number of steps between taken values. (default=None)
                include_final_step: Should the final step (the final value) also be included even if outside the stepping. (default=False)


    :return: Plotly figure object that can be displayed using display(fig_obj).
    """
    default_config = eu.AttrDict(

        # allows to display the moving average per element over n steps
        moving_average=eu.AttrDict(
            n=1,
            mode='fill_start'
        ),

        data_filter=eu.AttrDict(
            every_nth_step=eu.AttrDict(
                step=None,
                include_final_step=False),
        ),

        subplots=eu.AttrDict(  # paramters for the 'plotly.subplots.make_subplots' function
            rows=None,
            cols=None,
            print_grid=False
        ),
        std=eu.AttrDict(
            style='shaded',  # or errorbar
            steps=1,
            visible=True
        ),
        init_mode='mean_std',  # mean_std, mean, elements
        plotly_format='webgl',  # webgl or svg
        error_type='std', # std or sem (standard error of the mean)
        layout=eu.AttrDict(

            default_xaxis=eu.AttrDict(),
            # if several subplots, then these are the default values for all xaxis config in fig.layout
            default_yaxis=eu.AttrDict(),
            # if several subplots, then these are the default values for all yaxis config in fig.layout

            xaxis=eu.AttrDict(),
            yaxis=eu.AttrDict(),

            updatemenus=[
                eu.AttrDict(type="buttons",
                         active=0,
                         buttons=[
                             eu.AttrDict(label='mean + std',
                                      method='restyle',
                                      args=[{'visible': []}]),
                             eu.AttrDict(label='mean',
                                      method='restyle',
                                      args=[{'visible': []}]),
                             eu.AttrDict(label='elements',
                                      method='restyle',
                                      args=[{'visible': []}]),
                         ],
                         direction='right',
                         pad={'t': 70},
                         x=1,
                         xanchor='right',
                         y=0,
                         yanchor='top'
                         ),
            ]
        ),

        default_trace=eu.AttrDict(),

        default_mean_trace=eu.AttrDict(
            legendgroup='<subplot_idx>-<trace_idx>',  # subplot_idx, trace_idx
            hoverinfo='text+x',
        ),
        default_subplot_mean_traces=[],  # default config of traces per subplot
        mean_traces=[],

        default_std_trace=eu.AttrDict(
            legendgroup='<mean_trace_legendgroup>',  # subplot_idx, trace_idx, mean_trace_legendgroup
            hoverinfo='none',
            showlegend=False,
        ),
        default_subplot_std_traces=[],  # default config of traces per subplot
        std_traces=[],

        default_element_trace=eu.AttrDict(  # overall default
            legendgroup=None,
            # subplot_idx, trace_idx, elem_idx, subelem_idx, mean_trace_legendgroup, std_trace_legendgroup
        ),
        default_subplot_element_traces=[],  # default per subplot
        default_data_element_traces=[],  # default per data item
        element_traces=[],  # individual items

        # label configurations

        labels=[],  # holds all labels in a specific structure

        default_mean_label='<trace_idx>',
        mean_labels=[],

        default_element_label='<mean_label> - <subelem_idx>',
        # possible replacements: <mean_label>, <subelem_idx>, <elem_idx>, <trace_idx>
        element_labels=[],

        default_colors=plotly.colors.DEFAULT_PLOTLY_COLORS,
    )
    config = eu.combine_dicts(kwargs, config, default_config)

    if data is None:
        data = np.array([])

    # format data in form [subplot_idx:list][trace_idx:list][elems_per_trace:numpy.ndarray]
    if isinstance(data, np.ndarray):
        data = [[data]]
    elif isinstance(data, list) and isinstance(data[0], np.ndarray):
        data = [data]
    elif not isinstance(data, list) and not isinstance(data[0], list) and not isinstance(data[0][0], np.ndarray):
        raise ValueError('Unsupported type of data!')

    # handle different input formats of labels
    if config.labels:
        # if only labels for mean-traces are given, then add an empty label for the sub figure
        if isinstance(config.labels, list) and not isinstance(config.labels[0], tuple):
            config.labels = [('', config.labels)]

        # if no labels are given for elements, then create an empty list for element labels
        for ds_idx in range(len(config.labels)):
            for trace_idx in range(len(config.labels[ds_idx][1])):
                if not isinstance(config.labels[ds_idx][1][trace_idx], tuple):
                    config.labels[ds_idx][1][trace_idx] = (config.labels[ds_idx][1][trace_idx], [])

    # subplot_titles
    if config.labels and ('subplot_titles' not in config.subplots or config.subplots.subplot_titles == []):
        config.subplots.subplot_titles = [subplot_labels[0] for subplot_labels in config.labels]

    # identify the number of subplots
    n_subplots = len(data)

    # if not defined, set rows and cols of subplots
    if config.subplots.rows is None and config.subplots.cols is None:
        config.subplots.rows = n_subplots
        config.subplots.cols = 1
    elif config.subplots.rows is not None and config.subplots.cols is None:
        config.subplots.cols = int(np.ceil(n_subplots / config.subplots.rows))
    elif config.subplots.rows is None and config.subplots.cols is not None:
        config.subplots.rows = int(np.ceil(n_subplots / config.subplots.cols))

    if config.plotly_format.lower() == 'webgl':
        plotly_scatter_plotter = plotly.graph_objs.Scattergl
    elif config.plotly_format.lower() == 'svg':
        plotly_scatter_plotter = plotly.graph_objs.Scatter
    else:
        raise ValueError('Unknown config {!r} for plotly_format! Allowed values: \'webgl\', \'svg\'.')

    # make figure with subplots
    fig = plotly.subplots.make_subplots(**config.subplots)

    mean_traces = []
    elem_traces = []

    elem_idx = 0

    # interate over subplots
    for subplot_idx, subplot_data in enumerate(data):

        subplot_mean_traces = []
        subplot_elem_traces = []

        # iterate over traces
        for trace_idx, cur_data in enumerate(subplot_data):

            # do not plot data, if it does not exits
            if cur_data is not None:

                # in case the data is only 1 point, add an extra dimension
                if np.ndim(cur_data) == 1:
                    cur_data = np.array([cur_data]).transpose()

                # create a moving average over the data if requested
                if config.moving_average is not None and config.moving_average.n != 1:
                    cur_data = eu.misc.moving_average(
                        cur_data,
                        config.moving_average.n,
                        config.moving_average.mode)

                # define standard x_values
                x_values = list(range(cur_data.shape[1]))

                # filter the data if requested
                if config.data_filter.every_nth_step is not None:

                    if isinstance(config.data_filter.every_nth_step, dict):
                        step = config.data_filter.every_nth_step.step
                        is_include_final_step = config.data_filter.every_nth_step.include_final_step
                    else:
                        step = config.data_filter.every_nth_step
                        is_include_final_step = False

                    # detect if the final step was not in the selection
                    if is_include_final_step and cur_data.shape[1] % step == 0:
                        inds = np.zeros(cur_data.shape[1], dtype=bool)
                        inds[::step] = True
                        inds[-1] = True

                        cur_data = cur_data[:, inds]
                        x_values = x_values[::step] + [x_values[-1]]
                    else:
                        cur_data = cur_data[:, ::step]
                        x_values = x_values[::step]

                mean_data = np.nanmean(cur_data, axis=0)

                if config.error_type == 'std':
                    std_data = np.nanstd(cur_data, axis=0)
                elif config.error_type == 'sem':
                    std_data = np.nanstd(cur_data, axis=0, ddof=1) / np.sqrt(np.shape(cur_data)[0])
                else:
                    raise ValueError('Unknown error_type!')

                info_text = ['{} ± {}'.format(mean_data[idx], std_data[idx]) for idx in range(len(mean_data))]

                # define label of the trace
                if config.labels:
                    mean_label = config.labels[subplot_idx][1][trace_idx][0]
                else:
                    mean_label = config.default_mean_label
                    if len(config.mean_labels) > trace_idx:
                        mean_label = config.mean_labels[trace_idx]
                mean_label = eu.misc.replace_str_from_dict(str(mean_label), {'<trace_idx>': trace_idx})

                mean_trace_params = dict(
                    x=x_values,
                    y=mean_data,
                    line=dict(color=config.default_colors[trace_idx % len(config.default_colors)]),
                    name=mean_label,
                    text=info_text,
                )

                mean_trace_config = eu.combine_dicts(config.default_mean_trace, config.default_trace)
                if len(config.default_subplot_mean_traces) > subplot_idx:
                    mean_trace_config = eu.combine_dicts(config.default_subplot_mean_traces[subplot_idx], mean_trace_config)
                if len(config.mean_traces) > trace_idx:
                    mean_trace_config = eu.combine_dicts(config.mean_traces[trace_idx], mean_trace_config)

                mean_trace_params = eu.combine_dicts(mean_trace_config, mean_trace_params)

                # handle legendgroup
                mean_trace_legendgroup = mean_trace_params.legendgroup
                if isinstance(mean_trace_legendgroup, str):
                    mean_trace_legendgroup = eu.misc.replace_str_from_dict(mean_trace_legendgroup,
                                                                           {'<trace_idx>': trace_idx,
                                                                            '<subplot_idx>': subplot_idx})
                mean_trace_params.legendgroup = mean_trace_legendgroup

                cur_mean_trace = plotly_scatter_plotter(**mean_trace_params)
                subplot_mean_traces.append(cur_mean_trace)

                # handle trace for std values

                if config.std.style.lower() == 'shaded':

                    fill_color = config.default_colors[trace_idx % len(config.default_colors)]
                    fill_color = fill_color.replace('rgb', 'rgba')
                    fill_color = fill_color.replace(')', ', 0.2)')

                    std_trace_params = dict(
                        x=x_values + x_values[::-1],
                        y=np.hstack((mean_data + std_data, mean_data[::-1] - std_data[::-1])),
                        fill='tozerox',
                        line=dict(color='rgba(255,255,255,0)'),
                        fillcolor=fill_color,
                    )

                elif config.std.style.lower() == 'errorbar':

                    std_trace_params = dict(
                        x=x_values[::config.std.steps],
                        y=mean_data[::config.std.steps],
                        error_y=dict(type='data', array=std_data, visible=True),
                        mode='markers',
                        line=dict(color=config.default_colors[trace_idx % len(config.default_colors)]),
                        marker=dict(size=0, opacity=0),
                    )

                else:
                    raise ValueError(
                        'Unknown config.std.style ({!r})! Options: \'shaded\', \'errorbar\''.format(config.std.type))

                std_trace_config = eu.combine_dicts(config.default_std_trace, config.default_trace)
                if len(config.default_subplot_std_traces) > subplot_idx:
                    std_trace_config = eu.combine_dicts(config.default_subplot_std_traces[subplot_idx], std_trace_config)
                if len(config.std_traces) > trace_idx:
                    std_trace_config = eu.combine_dicts(config.std_traces[trace_idx], std_trace_config)
                std_trace_params = eu.combine_dicts(std_trace_config, std_trace_params)

                # handle legendgroup
                std_trace_legendgroup = std_trace_params.legendgroup
                if isinstance(std_trace_legendgroup, str):
                    std_trace_legendgroup = eu.misc.replace_str_from_dict(std_trace_legendgroup,
                                                                          {'<trace_idx>': trace_idx,
                                                                           '<subplot_idx>': subplot_idx,
                                                                           '<mean_trace_legendgroup>': mean_trace_legendgroup})
                std_trace_params.legendgroup = std_trace_legendgroup

                cur_std_trace = plotly_scatter_plotter(**std_trace_params)
                subplot_mean_traces.append(cur_std_trace)

                # traces for each data element
                n_elems = cur_data.shape[0]
                color_coeff_step = 1 / n_elems
                cur_color_coeff = 0 + color_coeff_step
                for cur_elem_idx in range(n_elems):

                    if config.labels and len(config.labels[subplot_idx][1][trace_idx][1]) > cur_elem_idx:
                        element_label = config.labels[subplot_idx][1][trace_idx][1][cur_elem_idx]
                    else:
                        element_label = config.default_element_label
                    if len(config.element_labels) > trace_idx:
                        element_label = config.element_labels[trace_idx]
                    element_label = eu.misc.replace_str_from_dict(str(element_label),
                                                                  {'<trace_idx>': trace_idx,
                                                                   '<subelem_idx>': cur_elem_idx,
                                                                   '<elem_idx>': elem_idx,
                                                                   '<mean_label>': mean_label})

                    color = eu.gui.misc.transform_color_str_to_tuple(
                        config.default_colors[trace_idx % len(config.default_colors)])
                    color = (color[0],
                             int(color[1] * cur_color_coeff),
                             int(color[2] * cur_color_coeff),
                             int(color[3] * cur_color_coeff))
                    color = eu.gui.misc.transform_color_tuple_to_str(color)
                    cur_color_coeff += color_coeff_step

                    element_trace_params = dict(
                        x=x_values,
                        y=cur_data[cur_elem_idx, :],
                        line=dict(color=color),
                        name=element_label,
                        visible=True,
                    )

                    element_trace_config = eu.combine_dicts(config.default_element_trace, config.default_trace)
                    if len(config.default_subplot_element_traces) > subplot_idx:
                        element_trace_config = eu.combine_dicts(config.default_subplot_element_traces[subplot_idx],
                                                                element_trace_config)
                    if len(config.default_data_element_traces) > cur_elem_idx:
                        element_trace_config = eu.combine_dicts(config.default_data_element_traces[cur_elem_idx],
                                                                element_trace_config)
                    if len(config.element_traces) > elem_idx:
                        element_trace_config = eu.combine_dicts(config.element_traces[elem_idx], element_trace_config)

                    element_trace_params = eu.combine_dicts(element_trace_config, element_trace_params)

                    # handle legendgroup
                    element_trace_legendgroup = element_trace_params.legendgroup
                    if isinstance(element_trace_legendgroup, str):
                        element_trace_legendgroup = eu.misc.replace_str_from_dict(
                            element_trace_legendgroup,
                            {'<subelem_idx>': cur_elem_idx,
                             '<elem_idx>': elem_idx,
                             '<trace_idx>': trace_idx,
                             '<subplot_idx>': subplot_idx,
                             '<mean_trace_legendgroup>': mean_trace_legendgroup,
                             '<std_trace_legendgroup>': std_trace_legendgroup})
                    element_trace_params.legendgroup = element_trace_legendgroup

                    cur_elem_trace = plotly_scatter_plotter(**element_trace_params)
                    subplot_elem_traces.append(cur_elem_trace)

                    elem_idx += 1

        mean_traces.append(subplot_mean_traces)
        elem_traces.append(subplot_elem_traces)

    # set for the std toggle buttons which traces should be hidden and which ones should be shown
    layout = config.layout

    # set default values for all layouts
    def set_axis_properties_by_default(axis_name, fig_layout, config_layout):
        # sets the axis properties to default values

        def set_single_axis_property_default(cur_axis_name, default_name):
            if cur_axis_name in fig_layout or cur_axis_name in config_layout:
                cur_config = config_layout[cur_axis_name] if cur_axis_name in config_layout else dict()
                config_layout[cur_axis_name] = eu.combine_dicts(cur_config, config_layout[default_name])

        default_name = 'default_' + axis_name

        set_single_axis_property_default(axis_name, default_name)
        set_single_axis_property_default(axis_name + '1', default_name)
        axis_idx = 2
        while True:
            cur_axis_name = axis_name + str(axis_idx)

            if cur_axis_name not in fig_layout and cur_axis_name not in config_layout:
                break

            set_single_axis_property_default(cur_axis_name, default_name)
            axis_idx += 1

    set_axis_properties_by_default('xaxis', fig['layout'], layout)
    set_axis_properties_by_default('yaxis', fig['layout'], layout)

    # remove default fields, because they are not true proerties of the plotly layout
    del (layout['default_xaxis'])
    del (layout['default_yaxis'])

    update_menus_visible_meanstd = []
    update_menus_visible_mean = []
    update_menus_visible_elements = []

    for subplot_idx in range(len(mean_traces)):
        update_menus_visible_meanstd.extend(
            [True, True] * int(len(mean_traces[subplot_idx]) / 2) + [False] * int(len(elem_traces[subplot_idx])))
        update_menus_visible_mean.extend(
            [True, False] * int(len(mean_traces[subplot_idx]) / 2) + [False] * int(len(elem_traces[subplot_idx])))

        element_default_visibility = [elem_trace['visible'] for elem_trace in elem_traces[subplot_idx]]
        update_menus_visible_elements.extend(
            [False, False] * int(len(mean_traces[subplot_idx]) / 2) + element_default_visibility)

    if layout.updatemenus:

        layout.updatemenus[0]['buttons'][0]['args'][0]['visible'] = update_menus_visible_meanstd
        layout.updatemenus[0]['buttons'][1]['args'][0]['visible'] = update_menus_visible_mean
        layout.updatemenus[0]['buttons'][2]['args'][0]['visible'] = update_menus_visible_elements

        if config.init_mode == 'mean_std':
            config.layout.updatemenus[0]['active'] = 0
        elif config.init_mode == 'mean':
            config.layout.updatemenus[0]['active'] = 1
        elif config.init_mode == 'elements':
            config.layout.updatemenus[0]['active'] = 2
        else:
            raise ValueError(
                'Value {!r} for \'config.init_mode\' is not supported! Only \'mean_std\',\'mean\',\'elements\'.'.format(
                    config.init_mode))

    if config.init_mode == 'mean_std':
        trace_visibility = update_menus_visible_meanstd
    elif config.init_mode == 'mean':
        trace_visibility = update_menus_visible_mean
    elif config.init_mode == 'elements':
        trace_visibility = update_menus_visible_elements
    else:
        raise ValueError(
            'Value {!r} for \'config.init_mode\' is not supported! Only \'mean_std\',\'mean\',\'elements\'.'.format(
                config.init_mode))

    cur_row = 1
    cur_col = 1
    for subplot_idx in range(n_subplots):

        n_traces = len(mean_traces[subplot_idx]) + len(elem_traces[subplot_idx])

        fig.add_traces(mean_traces[subplot_idx] + elem_traces[subplot_idx],
                       rows=[cur_row] * n_traces,
                       cols=[cur_col] * n_traces)

        if cur_col < config.subplots.cols:
            cur_col += 1
        else:
            cur_col = 1
            cur_row += 1

    for trace_idx in range(len(fig['data'])):
        fig['data'][trace_idx]['visible'] = trace_visibility[trace_idx]

    fig['layout'].update(layout)

    return fig



##
## This file is part of the exputils package.
##
## Copyright: INRIA
## Year: 2022
## Contact: chris.reinke@inria.fr
##
## exputils is provided under GPL-3.0-or-later
##
def transform_color_str_to_tuple(colors):
    is_input_list = isinstance(colors, list)

    if not is_input_list:
        colors = [colors]

    out_color = []
    for color in colors:
        col_elements = color.replace('(', ',').replace(')', ',').split(',')
        out_color.append((col_elements[0], int(col_elements[1]), int(col_elements[2]), int(col_elements[3])))

    if is_input_list:
        return out_color
    else:
        return out_color[0]


def transform_color_tuple_to_str(colors):
    is_input_list = isinstance(colors, list)

    if not is_input_list:
        colors = [colors]

    out_color = ['{}({}, {}, {})'.format(color[0], color[1], color[2], color[3]) for color in colors]

    if is_input_list:
        return out_color
    else:
        return out_color[0]



##
## This file is part of the exputils package.
##
## Copyright: INRIA
## Year: 2022
## Contact: chris.reinke@inria.fr
##
## exputils is provided under GPL-3.0-or-later
##

# This code was adapted from https://github.com/Infinidat/munch
#
# Copyright (c) 2010 David Schoonover
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
import exputils
import yaml
from collections import defaultdict
from collections.abc import Mapping
from six import iteritems, iterkeys  # pylint: disable=unused-import
from copy import deepcopy
try:
    import json
except ImportError:
    import simplejson as json

class AttrDict(dict):
    """ A dictionary that provides attribute-style access.
        >>> b = AttrDict()
        >>> b.hello = 'world'
        >>> b.hello
        'world'
        >>> b['hello'] += "!"
        >>> b.hello
        'world!'
        >>> b.foo = AttrDict(lol=True)
        >>> b.foo.lol
        True
        >>> b.foo is b['foo']
        True
        A AttrDict is a subclass of dict; it supports all the methods a dict does...
        >>> sorted(b.keys())
        ['foo', 'hello']
        Including update()...
        >>> b.update({ 'ponies': 'are pretty!' }, hello=42)
        >>> print (repr(b))
        AttrDict({'ponies': 'are pretty!', 'foo': Munch({'lol': True}), 'hello': 42})
        As well as iteration...
        >>> sorted([ (k,b[k]) for k in b ])
        [('foo', AttrDict({'lol': True})), ('hello', 42), ('ponies', 'are pretty!')]
        And "splats".
        >>> "The {knights} who say {ni}!".format(**AttrDict(knights='lolcats', ni='can haz'))
        'The lolcats who say can haz!'
        See unmunchify/AttrDict.toDict, munchify/AttrDict.fromDict for notes about conversion.
    """

    # only called if k not found in normal places
    def __getattr__(self, k):
        """ Gets key if it exists, otherwise throws AttributeError.
            nb. __getattr__ is only called if key is not found in normal places.
            >>> b = AttrDict(bar='baz', lol={})
            >>> b.foo
            Traceback (most recent call last):
                ...
            AttributeError: foo
            >>> b.bar
            'baz'
            >>> getattr(b, 'bar')
            'baz'
            >>> b['bar']
            'baz'
            >>> b.lol is b['lol']
            True
            >>> b.lol is getattr(b, 'lol')
            True
        """
        try:
            # Throws exception if not in prototype chain
            return object.__getattribute__(self, k)
        except AttributeError:
            try:
                return self[k]
            except KeyError:
                raise AttributeError(k)


    def __setattr__(self, k, v):
        """ Sets attribute k if it exists, otherwise sets key k. A KeyError
            raised by set-item (only likely if you subclass Munch) will
            propagate as an AttributeError instead.
            >>> b = AttrDict(foo='bar', this_is='useful when subclassing')
            >>> hasattr(b.values, '__call__')
            True
            >>> b.values = 'uh oh'
            >>> b.values
            'uh oh'
            >>> b['values']
            Traceback (most recent call last):
                ...
            KeyError: 'values'
        """
        try:
            # Throws exception if not in prototype chain
            object.__getattribute__(self, k)
        except AttributeError:
            try:
                self[k] = v
            except:
                raise AttributeError(k)
        else:
            object.__setattr__(self, k, v)


    def __delattr__(self, k):
        """ Deletes attribute k if it exists, otherwise deletes key k. A KeyError
            raised by deleting the key--such as when the key is missing--will
            propagate as an AttributeError instead.
            >>> b = AttrDict(lol=42)
            >>> del b.lol
            >>> b.lol
            Traceback (most recent call last):
                ...
            AttributeError: lol
        """
        try:
            # Throws exception if not in prototype chain
            object.__getattribute__(self, k)
        except AttributeError:
            try:
                del self[k]
            except KeyError:
                raise AttributeError(k)
        else:
            object.__delattr__(self, k)


    def toDict(self):
        """ Recursively converts a munch back into a dictionary.
            >>> b = AttrDict(foo=AttrDict(lol=True), hello=42, ponies='are pretty!')
            >>> sorted(b.toDict().items())
            [('foo', {'lol': True}), ('hello', 42), ('ponies', 'are pretty!')]
            See unmunchify for more info.
        """
        return attrdict_to_dict(self)


    @property
    def __dict__(self):
        return self.toDict()


    def __repr__(self):
        """ Invertible* string-form of a Munch.
            >>> b = AttrDict(foo=AttrDict(lol=True), hello=42, ponies='are pretty!')
            >>> print (repr(b))
            Munch({'ponies': 'are pretty!', 'foo': Munch({'lol': True}), 'hello': 42})
            >>> eval(repr(b))
            Munch({'ponies': 'are pretty!', 'foo': Munch({'lol': True}), 'hello': 42})
            >>> with_spaces = AttrDict({1: 2, 'a b': 9, 'c': AttrDict({'simple': 5})})
            >>> print (repr(with_spaces))
            Munch({'a b': 9, 1: 2, 'c': Munch({'simple': 5})})
            >>> eval(repr(with_spaces))
            Munch({'a b': 9, 1: 2, 'c': Munch({'simple': 5})})
            (*) Invertible so long as collection contents are each repr-invertible.
        """
        return '{0}({1})'.format(self.__class__.__name__, dict.__repr__(self))


    def __dir__(self):
        return list(iterkeys(self))


    def __getstate__(self):
        """ Implement a serializable interface used for pickling.
        See https://docs.python.org/3.6/library/pickle.html.
        """
        return {k: v for k, v in self.items()}


    def __setstate__(self, state):
        """ Implement a serializable interface used for pickling.
        See https://docs.python.org/3.6/library/pickle.html.
        """
        self.clear()
        self.update(state)


    __members__ = __dir__  # for python2.x compatibility


    def __eq__(self, other):
        '''Is the dict equal to another dict. Allows to compare numpy arrays.'''
        return exputils.misc.dict_equal(self, other)


    @classmethod
    def from_dict(cls, d):
        """ Recursively transforms a dictionary into a AttrDict via copy.
            >>> b = AttrDict.from_dict({'urmom': {'sez': {'what': 'what'}}})
            >>> b.urmom.sez.what
            'what'
            See dict_to_attrdict for more info.
        """
        return dict_to_attrdict(d, cls)


    def copy(self):
        return type(self).from_dict(self)


    def to_json(self, **options):
        """ Serializes this AttrDict to JSON. Accepts the same keyword options as `json.dumps()`.
            >>> b = AttrDict(foo=AttrDict(lol=True), hello=42, ponies='are pretty!')
            >>> json.dumps(b) == b.to_json()
            True

            Allows to dump numpy arrays into JSON.
        """

        # allow to dump numpy into json
        if 'cls' not in options:
            options['cls'] = exputils.io.json.ExputilsJSONEncoder

        return json.dumps(self, **options)


    @classmethod
    def from_json(cls, json_data, is_transform_ints=True, **options):
        """ Loads an AttrDict from JSON. Accepts the same keyword options as `json.loads()`."""

        # allow to load numpy from json
        if 'cls' not in options:
            options['object_hook'] = exputils.io.json.exputils_json_object_hook

        loaded_json = json.loads(json_data, **options)

        # chenge possible int keys to ints as json changes ints to strings
        if is_transform_ints:
            loaded_json = exputils.io.convert_json_dict_keys_to_ints(loaded_json)

        return dict_to_attrdict(loaded_json, cls)


    def to_json_file(self, filepath, **options):
        exputils.io.save_dict_as_json_file(self, filepath, **options)


    @classmethod
    def from_json_file(cls, filepath, **options):
        loaded_dict = exputils.io.load_dict_from_json_file(filepath, **options)
        return dict_to_attrdict(loaded_dict, cls)


    def to_yaml(self, path) -> None:

        with open(path, 'w') as output_file:
            yaml.dump(self.toDict(), output_file)

    @classmethod
    def from_yaml(cls, filepath, **options):
        with open(filepath, 'r') as config_file:
            loaded_dict = yaml.load(config_file.read(),
                                    yaml.FullLoader)
        return dict_to_attrdict(loaded_dict, cls)


class AutoAttrDict(AttrDict):
    def __setattr__(self, k, v):
        """ Works the same as AttrDict.__setattr__ but if you supply
            a dictionary as value it will convert it to another AttrDict.
        """
        if isinstance(v, dict) and not isinstance(v, (AutoAttrDict, AttrDict)):
            v = dict_to_attrdict(v, AutoAttrDict)
        super(AutoAttrDict, self).__setattr__(k, v)


class DefaultAttrDict(AttrDict):
    """
    A AttrDict that returns a user-specified value for missing keys.
    """


    def __init__(self, *args, **kwargs):
        """ Construct a new DefaultAttrDict. Like collections.defaultdict, the
            first argument is the default value; subsequent arguments are the
            same as those for dict.
        """
        # Mimic collections.defaultdict constructor
        if args:
            default = args[0]
            args = args[1:]
        else:
            default = None
        super(DefaultAttrDict, self).__init__(*args, **kwargs)
        self.__default__ = default


    def __getattr__(self, k):
        """ Gets key if it exists, otherwise returns the default value."""
        try:
            return super(DefaultAttrDict, self).__getattr__(k)
        except AttributeError:
            return self.__default__


    def __setattr__(self, k, v):
        if k == '__default__':
            object.__setattr__(self, k, v)
        else:
            return super(DefaultAttrDict, self).__setattr__(k, v)


    def __getitem__(self, k):
        """ Gets key if it exists, otherwise returns the default value."""
        try:
            return super(DefaultAttrDict, self).__getitem__(k)
        except KeyError:
            return self.__default__


    def __getstate__(self):
        """ Implement a serializable interface used for pickling.
        See https://docs.python.org/3.6/library/pickle.html.
        """
        return (self.__default__, {k: v for k, v in self.items()})


    def __setstate__(self, state):
        """ Implement a serializable interface used for pickling.
        See https://docs.python.org/3.6/library/pickle.html.
        """
        self.clear()
        default, state_dict = state
        self.update(state_dict)
        self.__default__ = default


    @classmethod
    def from_dict(cls, d, default=None):
        # pylint: disable=arguments-differ
        return dict_to_attrdict(d, factory=lambda d_: cls(default, d_))


    def copy(self):
        return type(self).from_dict(self, default=self.__default__)


    def __repr__(self):
        return '{0}({1!r}, {2})'.format(
            type(self).__name__, self.__undefined__, dict.__repr__(self))


class DefaultFactoryAttrDict(defaultdict, AttrDict):
    """ A AttrDict that calls a user-specified function to generate values for
        missing keys like collections.defaultdict.
        >>> b = DefaultFactoryAttrDict(list, {'hello': 'world!'})
        >>> b.hello
        'world!'
        >>> b.foo
        []
        >>> b.bar.append('hello')
        >>> b.bar
        ['hello']
    """


    def __init__(self, default_factory, *args, **kwargs):
        # pylint: disable=useless-super-delegation
        super(DefaultFactoryAttrDict, self).__init__(default_factory, *args, **kwargs)


    @classmethod
    def from_dict(cls, d, default_factory):
        # pylint: disable=arguments-differ
        return dict_to_attrdict(d, factory=lambda d_: cls(default_factory, d_))


    def copy(self):
        return type(self).from_dict(self, default_factory=self.default_factory)


    def __repr__(self):
        factory = self.default_factory.__name__
        return '{0}({1}, {2})'.format(
            type(self).__name__, factory, dict.__repr__(self))


# While we could convert abstract types like Mapping or Iterable, I think
# munchify is more likely to "do what you mean" if it is conservative about
# casting (ex: isinstance(str,Iterable) == True ).
#
# Should you disagree, it is not difficult to duplicate this function with
# more aggressive coercion to suit your own purposes.

def dict_to_attrdict(x, factory=AttrDict):
    """ Recursively transforms a dictionary into a AttrDict via copy.
        >>> b = dict_to_attrdict({'urmom': {'sez': {'what': 'what'}}})
        >>> b.urmom.sez.what
        'what'
        munchify can handle intermediary dicts, lists and tuples (as well as
        their subclasses), but ymmv on custom datatypes.
        >>> b = dict_to_attrdict({ 'lol': ('cats', {'hah':'i win again'}),
        ...         'hello': [{'french':'salut', 'german':'hallo'}] })
        >>> b.hello[0].french
        'salut'
        >>> b.lol[1].hah
        'i win again'
        nb. As dicts are not hashable, they cannot be nested in sets/frozensets.
    """
    if isinstance(x, dict):
        return factory((k, dict_to_attrdict(v, factory)) for k, v in iteritems(x))
    elif isinstance(x, (list, tuple)):
        return type(x)(dict_to_attrdict(v, factory) for v in x)
    else:
        return x


def attrdict_to_dict(x):
    """ Recursively converts a AttrDict into a dictionary.
        >>> b = AttrDict(foo=AttrDict(lol=True), hello=42, ponies='are pretty!')
        >>> sorted(attrdict_to_dict(b).items())
        [('foo', {'lol': True}), ('hello', 42), ('ponies', 'are pretty!')]
        unmunchify will handle intermediary dicts, lists and tuples (as well as
        their subclasses), but ymmv on custom datatypes.
        >>> b = AttrDict(foo=['bar', AttrDict(lol=True)], hello=42,
        ...         ponies=('are pretty!', AttrDict(lies='are trouble!')))
        >>> sorted(attrdict_to_dict(b).items()) #doctest: +NORMALIZE_WHITESPACE
        [('foo', ['bar', {'lol': True}]), ('hello', 42), ('ponies', ('are pretty!', {'lies': 'are trouble!'}))]
        nb. As dicts are not hashable, they cannot be nested in sets/frozensets.
    """
    if isinstance(x, dict):
        return dict((k, attrdict_to_dict(v)) for k, v in iteritems(x))
    elif isinstance(x, (list, tuple)):
        return type(x)(attrdict_to_dict(v) for v in x)
    else:
        return x


# Serialization

# try:
#     # Attempt to register ourself with PyYAML as a representer
#     import yaml
#     from yaml.representer import Representer, SafeRepresenter
#
#
#     def from_yaml(loader, node):
#         """ PyYAML support for Munches using the tag `!munch` and `!munch.Munch`.
#             >>> import yaml
#             >>> yaml.load('''
#             ... Flow style: !munch.Munch { Clark: Evans, Brian: Ingerson, Oren: Ben-Kiki }
#             ... Block style: !munch
#             ...   Clark : Evans
#             ...   Brian : Ingerson
#             ...   Oren  : Ben-Kiki
#             ... ''') #doctest: +NORMALIZE_WHITESPACE
#             {'Flow style': Munch(Brian='Ingerson', Clark='Evans', Oren='Ben-Kiki'),
#              'Block style': Munch(Brian='Ingerson', Clark='Evans', Oren='Ben-Kiki')}
#             This module registers itself automatically to cover both Munch and any
#             subclasses. Should you want to customize the representation of a subclass,
#             simply register it with PyYAML yourself.
#         """
#         data = Munch()
#         yield data
#         value = loader.construct_mapping(node)
#         data.update(value)
#
#
#     def to_yaml_safe(dumper, data):
#         """ Converts Munch to a normal mapping node, making it appear as a
#             dict in the YAML output.
#             >>> b = Munch(foo=['bar', Munch(lol=True)], hello=42)
#             >>> import yaml
#             >>> yaml.safe_dump(b, default_flow_style=True)
#             '{foo: [bar, {lol: true}], hello: 42}\\n'
#         """
#         return dumper.represent_dict(data)
#
#
#     def to_yaml(dumper, data):
#         """ Converts Munch to a representation node.
#             >>> b = Munch(foo=['bar', Munch(lol=True)], hello=42)
#             >>> import yaml
#             >>> yaml.dump(b, default_flow_style=True)
#             '!munch.Munch {foo: [bar, !munch.Munch {lol: true}], hello: 42}\\n'
#         """
#         return dumper.represent_mapping(u('!munch.Munch'), data)
#
#
#     yaml.add_constructor(u('!munch'), from_yaml)
#     yaml.add_constructor(u('!munch.Munch'), from_yaml)
#
#     SafeRepresenter.add_representer(Munch, to_yaml_safe)
#     SafeRepresenter.add_multi_representer(Munch, to_yaml_safe)
#
#     Representer.add_representer(Munch, to_yaml)
#     Representer.add_multi_representer(Munch, to_yaml)
#
#
#     # Instance methods for YAML conversion
#     def toYAML(self, **options):
#         """ Serializes this Munch to YAML, using `yaml.safe_dump()` if
#             no `Dumper` is provided. See the PyYAML documentation for more info.
#             >>> b = Munch(foo=['bar', Munch(lol=True)], hello=42)
#             >>> import yaml
#             >>> yaml.safe_dump(b, default_flow_style=True)
#             '{foo: [bar, {lol: true}], hello: 42}\\n'
#             >>> b.toYAML(default_flow_style=True)
#             '{foo: [bar, {lol: true}], hello: 42}\\n'
#             >>> yaml.dump(b, default_flow_style=True)
#             '!munch.Munch {foo: [bar, !munch.Munch {lol: true}], hello: 42}\\n'
#             >>> b.toYAML(Dumper=yaml.Dumper, default_flow_style=True)
#             '!munch.Munch {foo: [bar, !munch.Munch {lol: true}], hello: 42}\\n'
#         """
#         opts = dict(indent=4, default_flow_style=False)
#         opts.update(options)
#         if 'Dumper' not in opts:
#             return yaml.safe_dump(self, **opts)
#         else:
#             return yaml.dump(self, **opts)
#
#
#     def fromYAML(*args, **kwargs):
#         return munchify(yaml.load(*args, **kwargs))
#
#
#     Munch.toYAML = toYAML
#     Munch.fromYAML = staticmethod(fromYAML)
#
# except ImportError:
#     pass


def combine_dicts(*args, is_recursive=True, copy_mode='deepcopy'):
    '''
    Combines several AttrDicts recursively.
    Kepps the values of the

    :param args:
    :param is_recursive: Should
    :param copy_mode: Defines how the dictionaries should be copied ('deepcopy', 'copy', 'none').
    :return: Combined AttrDict.
    '''

    args = list(args)

    # convert arguments to AttrDicts if they are not
    for idx in range(len(args)):
        if args[idx] is None:
            args[idx] = AttrDict()
        elif not isinstance(args[idx], AttrDict):
            args[idx] = AttrDict.from_dict(args[idx])

    # copy the dictionaries according to copy mode
    dicts = []
    for dict in args:
        if copy_mode.lower() == 'deepcopy':
            dicts.append(deepcopy(dict))
        elif copy_mode.lower() == 'copy':
            dicts.append(dict.copy())
        elif copy_mode is None or copy_mode.lower() == 'none':
            dicts.append(dict)
        else:
            raise ValueError('Unknown copy mode {!r}!'.format(copy_mode))

    # combine the dicts going from last to first
    for dict_idx in range(len(args)-1, 0, -1):

        for def_key, def_item in dicts[dict_idx].items():

            if not def_key in dicts[dict_idx-1]:
                # add default item if not found target
                dicts[dict_idx - 1][def_key] = def_item
            elif (is_recursive
                  and isinstance(def_item, Mapping)
                  and isinstance(dicts[dict_idx - 1][def_key], Mapping)):
                # If the value is a dictionary in the default and the target, then also set default
                # values for it.
                dicts[dict_idx - 1][def_key] = combine_dicts(dicts[dict_idx - 1][def_key],
                                                             def_item,
                                                             is_recursive=is_recursive,
                                                             copy_mode=copy_mode)

    return dicts[0]



##
## This file is part of the exputils package.
##
## Copyright: INRIA
## Year: 2022
## Contact: chris.reinke@inria.fr
##
## exputils is provided under GPL-3.0-or-later
##
import ipywidgets
import exputils as eu

#TODO: Feature - events for change of values

class TextSelectionAccordionWidget(ipywidgets.Accordion):

    @staticmethod
    def default_config():
        dc = eu.AttrDict()

        dc.title = '<selection>'

        dc.main_accordion = eu.AttrDict(
            #layout=eu.AttrDict(width='100%')
        )

        dc.main_vbox = eu.AttrDict(
            layout=eu.AttrDict(width='100%')
        )

        dc.default_selection_element_hbox = eu.AttrDict(
            layout=eu.AttrDict(width='100%')
        )

        dc.default_selection_element_label = eu.AttrDict(
            layout=eu.AttrDict(width='25%')
        )

        dc.default_selection_element_text = eu.AttrDict(
            layout=eu.AttrDict(width='75%')
        )

        dc.selection_elements = []

        return dc

    def __init__(self, config=None, **kwargs):
        self.config = eu.combine_dicts(kwargs, config, self.default_config())

        super().__init__( **self.config.main_accordion)

        # create the selection elements

        self.selection_text_widgets = dict()

        selection_hbox_widgets = []

        # create selection_elements as list of AttrDicts to have a consitent datatype
        self.selection_elements = []
        for elem_descr in self.config.selection_elements:

            if isinstance(elem_descr, str):
                elem_name = elem_descr
                elem_label = elem_descr
                elem_title_label = elem_descr
            else:
                elem_name = elem_descr['name']
                elem_label = elem_descr.get('label', elem_name)
                elem_title_label = elem_descr.get('title_label', elem_label)

            self.selection_elements.append(eu.AttrDict(
                name=elem_name,
                label = elem_label,
                title_label = elem_title_label))

        # create the gui elements
        for elem_descr in self.selection_elements:

            label = ipywidgets.Label(
                value=elem_descr.label,
                **self.config.default_selection_element_label
            )

            text_widget = ipywidgets.Text(
                value='',
                **self.config.default_selection_element_text
            )

            hbox = ipywidgets.HBox(
                children=[label, text_widget],
                **self.config.default_selection_element_hbox
            )

            selection_hbox_widgets.append(hbox)
            self.selection_text_widgets[elem_descr.name] = text_widget

            text_widget.observe(
                self._on_text_value_changed,
                names='value')

        main_vbox = ipywidgets.VBox(
            children=selection_hbox_widgets,
            **self.config.main_vbox
        )

        eu.gui.jupyter.add_children_to_widget(self, main_vbox)


    @property
    def selection(self):
        selection = eu.AttrDict()
        for elem_descr in self.selection_elements:
            selection[elem_descr.name] = self.selection_text_widgets[elem_descr.name].value
        return selection


    @selection.setter
    def selection(self, selection):
        for elem_name, elem_value in selection.items():
            self.selection_text_widgets[elem_name].value = elem_value


    def _on_text_value_changed(self, event_descr):
        self._update_title()


    def _update_title(self):

        if self.config.title is None:
            title_str = ''
        else:
            title_str = self.config.title

            # create selection title
            if '<selection>' in title_str:

                selection_str_elements = []
                for elem_descr in self.selection_elements:

                    if elem_descr.title_label.endswith(':'):
                        delimiter = ''
                    else:
                        delimiter = ':'

                    selection_str_elements.append('{}{} {}'.format(
                        elem_descr.title_label,
                        delimiter,
                        self.selection_text_widgets[elem_descr.name].value))

                title_str = title_str.replace(
                    '<selection>',
                    ', '.join(selection_str_elements))

        self.set_title(0, title_str)


##
## This file is part of the exputils package.
##
## Copyright: INRIA
## Year: 2022
## Contact: chris.reinke@inria.fr
##
## exputils is provided under GPL-3.0-or-later
##
import exputils.data
import exputils.gui
import exputils.io
import exputils.manage
import exputils.misc

from exputils.misc.attrdict import AttrDict
from exputils.misc.attrdict import AutoAttrDict
from exputils.misc.attrdict import DefaultAttrDict
from exputils.misc.attrdict import DefaultFactoryAttrDict
from exputils.misc.attrdict import combine_dicts

from exputils.misc import update_status

__version__ = '0.3.1'

DEFAULT_ODS_CONFIGURATION_FILE = 'experiment_configurations.ods'
DEFAULT_EXPERIMENTS_DIRECTORY = 'experiments'
EXPERIMENT_DIRECTORY_TEMPLATE = 'experiment_{:06d}'
REPETITION_DIRECTORY_TEMPLATE = 'repetition_{:06d}'
DEFAULT_DATA_DIRECTORY = 'data'  # name of the data directory under the experiments and repetition folders

REPETITION_DATA_KEY = 'repetition_data'  # key name  for repetition data in the experiment_data dictionary



##
## This file is part of the exputils package.
##
## Copyright: INRIA
## Year: 2022
## Contact: chris.reinke@inria.fr
##
## exputils is provided under GPL-3.0-or-later
##
import exputils as eu
import ipywidgets
import pandas as pd
import qgrid
import os
from exputils.gui.jupyter.base_widget import BaseWidget
import IPython
import warnings

# TODO: Feature - allow to filter datasources that should be loaded
# TODO: Feature - progress bar during data loading
# TODO: Bugfix - If the order of the items is changed rapidely (buttons quicly pressed) then sometimes a wrong item gets selected afterwards
#                It is unclear what causes this issue.

class ExperimentDataLoaderWidget(BaseWidget, ipywidgets.VBox):

    @staticmethod
    def default_config():
        dc = BaseWidget.default_config()

        dc.load_experiment_descriptions_function = eu.AttrDict(
            func=eu.data.load_experiment_descriptions
        )
        dc.load_experiment_data_function = eu.AttrDict(
            func=eu.data.load_experiment_data
        )
        dc.experiments_directory = os.path.join('..', eu.DEFAULT_EXPERIMENTS_DIRECTORY)

        dc.main_box = eu.AttrDict(
            layout=eu.AttrDict(
                width='99%',
                display='flex',
                flex_flow='column',
                align_items='stretch'))

        dc.top_button_box = eu.AttrDict(
            layout=eu.AttrDict(
                width='100%',
                display='flex',
                flex_flow='row',
                align_items='stretch'))

        dc.load_descr_button = eu.AttrDict(
            layout=eu.AttrDict(
                width = '75%',
                height = 'auto'),
            description = 'Update Descriptions',
            disabled = False,
            button_style = '',  # 'success', 'info', 'warning', 'danger' or ''
            tooltip = 'Update for the selected experiment and repetition.')

        dc.reset_descr_button = eu.AttrDict(
            layout=eu.AttrDict(
                width = '25%',
                height = 'auto'),
            description = 'Reset Descriptions',
            disabled = False,
            button_style = '',  # 'success', 'info', 'warning', 'danger' or ''
            tooltip = 'Reset all experiment descriptions.')

        dc.move_buttons_box = eu.AttrDict(
            layout=eu.AttrDict(
                width='100%',
                display='flex',
                flex_flow='row',
                align_items='stretch'))

        dc.move_up_button = eu.AttrDict(
            layout=eu.AttrDict(
                width = '25%',
                height = 'auto'),
            description = u'\u02C5', # 'down',
            disabled = False,
            button_style = '',  # 'success', 'info', 'warning', 'danger' or ''
            tooltip = 'Moves the selected experiments up in the order.  (Only works if data is not filtered.)')

        dc.move_down_button = eu.AttrDict(
            layout=eu.AttrDict(
                width = '25%',
                height = 'auto'),
            description = u'\u02C4', #'up',
            disabled = False,
            button_style = '',  # 'success', 'info', 'warning', 'danger' or ''
            tooltip = 'Moves the selected experiments down in the order.  (Only works if data is not filtered.)')

        dc.sort_by_id_button = eu.AttrDict(
            layout=eu.AttrDict(
                width='50%',
                height='auto'),
            description='Sort by Experiment ID',  # 'up',
            disabled=False,
            button_style='',  # 'success', 'info', 'warning', 'danger' or ''
            tooltip='Resorts the experiments according to their ID.')

        dc.data_buttons_box = eu.AttrDict(
            layout=eu.AttrDict(
                width='100%',
                display='flex',
                flex_flow='row',
                align_items='stretch'))

        dc.load_data_button = eu.AttrDict(
            layout=eu.AttrDict(
                width = '75%',
                height = 'auto'),
            description = 'Load Data',
            disabled = False,
            button_style = '',  # 'success', 'info', 'warning', 'danger' or ''
            tooltip = 'Load experimental data.')

        dc.empty_data_button = eu.AttrDict(
            layout=eu.AttrDict(
                width = '25%',
                height = 'auto'),
            description = 'Empty Data',
            disabled = False,
            button_style = '',  # 'success', 'info', 'warning', 'danger' or ''
            tooltip = 'Empties the loaded experimental data to free memory.')


        # naming of columns in the dataframe (key: name in experiment_description dict, value: name in dataframe)
        dc.dataframe_column_names = {'id': 'experiment id',
                                     'order': 'order',
                                     'is_load_data': 'load data',
                                     'short_name': 'short name',
                                     'name': 'name',
                                     'description': 'description',
                                     'directory': 'directory'}

        dc.qgrid_widget = eu.AttrDict(
            show_toolbar = True,
            grid_options = {'autoEdit': True,
                            'sortable': False},
            column_options = {'editable': False},
            column_definitions = {
                'load data': {'editable': True},
                'short name': {'editable': True},
                'name': {'editable': True},
                'description': {'editable': True}})

        dc.output_widget = eu.AttrDict()

        return dc


    def __init__(self, config=None, **kwargs):
        # constructor of BaseWidget
        super().__init__(config=config, **kwargs)
        # constructor of GridspecLayout
        super(BaseWidget, self).__init__(
            **self.config.main_box)

        self.experiment_descriptions = None
        self.experiment_data = None

        # list with registered event handlers for the data collected event
        self._on_experiment_data_loaded_event_handlers = []
        self._on_experiment_descriptions_updated_event_handlers = []

        self.load_state_backup()

        self.update_experiment_descriptions()

        # create gui elements
        self.load_descr_btn = ipywidgets.Button(**self.config.load_descr_button)
        self.reset_descr_btn = ipywidgets.Button(**self.config.reset_descr_button)
        self.top_button_box = ipywidgets.Box(
            children=[self.load_descr_btn, self.reset_descr_btn],
            **self.config.top_button_box)

        self.qgrid_widget = ipywidgets.Box()  # initialize with dummy, will be overridden by update function

        self.move_up_btn = ipywidgets.Button(**self.config.move_up_button)
        self.move_down_btn = ipywidgets.Button(**self.config.move_down_button)
        self.sort_by_id_button = ipywidgets.Button(**self.config.sort_by_id_button)
        self.move_buttons_box = ipywidgets.Box(
            children=[self.move_down_btn, self.move_up_btn, self.sort_by_id_button],
            **self.config.move_buttons_box)

        self.load_data_btn = ipywidgets.Button(**self.config.load_data_button)
        self.empty_data_btn = ipywidgets.Button(**self.config.empty_data_button)
        self.data_buttons_box = ipywidgets.Box(
            children=[self.load_data_btn, self.empty_data_btn],
            **self.config.data_buttons_box)

        eu.gui.jupyter.add_children_to_widget(
            self,
            [self.top_button_box, self.qgrid_widget, self.move_buttons_box, self.data_buttons_box])

        # create an output widget
        self._output_widget = None

        self._update_qgrid()

        # register events
        self.load_descr_btn.on_click(self._handle_load_descr_button_on_click)
        self.reset_descr_btn.on_click(self._handle_reset_descr_button_on_click)
        self.load_data_btn.on_click(self._handle_load_data_button_on_click)
        self.empty_data_btn.on_click(self._handle_empty_data_button_on_click)
        self.move_up_btn.on_click(self._handle_move_up_button_on_click)
        self.move_down_btn.on_click(self._handle_move_down_button_on_click)
        self.sort_by_id_button.on_click(self._handle_sort_by_id_button_on_click)

        self._handle_qgrid_cell_edited_is_active = True

    def _prepare_output_widget(self):

        if self._output_widget is None:
            self._output_widget = ipywidgets.Output(**self.config.output_widget)
            IPython.display.display(self._output_widget)
        else:
            warnings.resetwarnings()
            self._output_widget.clear_output(wait=False)

        return self._output_widget


    def _handle_load_descr_button_on_click(self, btn):
        # errors are plotted in output widget and it will be cleaned after next button press
        with self._prepare_output_widget():
            self.update_experiment_descriptions(is_reset=False)
            self._update_qgrid()


    def _handle_reset_descr_button_on_click(self, btn):
        # errors are plotted in output widget and it will be cleaned after next button press
        with self._prepare_output_widget():
            self.update_experiment_descriptions(is_reset=True)
            self._update_qgrid()


    def _handle_load_data_button_on_click(self, btn):
        # errors are plotted in output widget and it will be cleaned after next button press
        with self._prepare_output_widget():
            # load data and save widget state
            print('Load data ...')
            self.load_data()
            self.backup_state()
            print('Data successfully loaded.')


    def _handle_empty_data_button_on_click(self, btn):
        # errors are plotted in output widget and it will be cleaned after next button press
        with self._prepare_output_widget():
            # empty data and save widget state
            self.empty_data()
            self.backup_state()
            print('Emptied data.')


    def _handle_move_up_button_on_click(self, btn):
        # errors are plotted in output widget and it will be cleaned after next button press
        with self._prepare_output_widget():
            try:
                self.move_up_btn.disabled = True
                self.move_down_btn.disabled = True

                self.move_experiments_up()
            finally:
                self.move_up_btn.disabled = False
                self.move_down_btn.disabled = False

    def _handle_move_down_button_on_click(self, btn):
        # errors are plotted in output widget and it will be cleaned after next button press
        with self._prepare_output_widget():
            try:
                self.move_up_btn.disabled = True
                self.move_down_btn.disabled = True

                self.move_experiments_down()
            finally:
                self.move_up_btn.disabled = False
                self.move_down_btn.disabled = False

    def _handle_sort_by_id_button_on_click(self, btn):
        # errors are plotted in output widget and it will be cleaned after next button press
        with self._prepare_output_widget():
            self.resort_experiments_by_id()

    def _handle_qgrid_cell_edited(self, event, widget):
        with self._prepare_output_widget():

            if self._handle_qgrid_cell_edited_is_active:

                # update the experiment_description
                if event['name'] == 'cell_edited':

                    for expdescr_prop_name, df_col_name in self.config.dataframe_column_names.items():
                        if df_col_name == event['column']:
                            self.experiment_descriptions[event['index']][expdescr_prop_name] = event['new']
                            break

                    self.backup_state()

                    self._call_experiment_descriptions_updated_event()


    def _handle_qgrid_filter_changed(self, event, widget):
        with self._prepare_output_widget():

            # identify if a filter is active or not
            is_filter_active = len(self.qgrid_widget.df) != len(self.qgrid_widget.get_changed_df())

            if is_filter_active:
                # do not allow to change order
                self.move_up_btn.disabled = True
                self.move_down_btn.disabled = True

            else:
                # allow to change order
                self.move_up_btn.disabled = False
                self.move_down_btn.disabled = False

    def _update_qgrid(self):

        # convert experiment description to the dataframe
        df = pd.DataFrame()
        for exp_descr_field_name, df_column_name in self.config.dataframe_column_names.items():
            df[df_column_name] = [descr[exp_descr_field_name] for descr in self.experiment_descriptions.values()]

        df = df.set_index(self.config.dataframe_column_names['id'])

        # create a new qgrid widget with the dataframe
        for opt_name, opt_value in self.config.qgrid_widget.grid_options.items():
            qgrid.set_grid_option(opt_name, opt_value)

        self.qgrid_widget = qgrid.show_grid(df,
                                            column_options=self.config.qgrid_widget.column_options,
                                            column_definitions=self.config.qgrid_widget.column_definitions,
                                            show_toolbar=self.config.qgrid_widget.show_toolbar)

        eu.gui.jupyter.remove_children_from_widget(self, 1)
        eu.gui.jupyter.add_children_to_widget(self, self.qgrid_widget, idx=1)

        self.qgrid_widget.on('cell_edited', self._handle_qgrid_cell_edited)

        self.qgrid_widget.on('filter_changed', self._handle_qgrid_filter_changed)

        self.sort_grid_by_order()


    def sort_grid_by_order(self):
        # hack to resort the experiments in the grid according to the order field
        content = dict(
            type='change_sort',
            sort_field=self.config.dataframe_column_names['order'],
            sort_ascending=True)
        self.qgrid_widget._handle_qgrid_msg_helper(content)


    def on_experiment_descriptions_updated(self, handler):
        """
        Register an event handler for the case that the experiment descriptions was changed.
        Please note, that this does not mean that the data was loaded according to the new experiment descriptions.
        Use the on_experiment_data_loaded for this purpose.
        The handler receives a dict with information about the event.
        """
        self._on_experiment_descriptions_updated_event_handlers.append(handler)


    def _call_experiment_descriptions_updated_event(self):
        for handler in self._on_experiment_descriptions_updated_event_handlers:
            handler(eu.AttrDict(
                name='experiment_descriptions_updated',
                new=self.experiment_descriptions,
                owner=self,
                type='change'))


    def on_experiment_data_loaded(self, handler):
        """
        Register an event handler for the case new data was loaded.
        The handler receives a dict with information about the event.
        """
        self._on_experiment_data_loaded_event_handlers.append(handler)


    def _call_experiment_data_loaded_event(self):
        for handler in self._on_experiment_data_loaded_event_handlers:
            handler(eu.AttrDict(
                name='data_loaded',
                new=self.experiment_data,
                owner=self,
                type='change'))


    def move_experiments_up(self, selected_items=None, is_select_changed_items=True):

        try:
            self._handle_qgrid_cell_edited_is_active = False

            if selected_items is None:
                selected_items = self.qgrid_widget.get_selected_df()

            if len(selected_items) > 0:

                order_col_name = self.config.dataframe_column_names['order']
                experiment_id_col_name = self.config.dataframe_column_names['id']

                # select according to the order
                selected_items = selected_items.sort_values(by=order_col_name)
                qgrid_df = self.qgrid_widget.get_changed_df()

                # make order the index to allow to seach by it
                all_items_sorted_by_order = qgrid_df.sort_values(by=order_col_name).reset_index().set_index(order_col_name)

                selected_idx = len(selected_items) - 1
                while selected_idx >= 0:

                    # identify current block of consequetively selected items
                    current_block_exp_ids = []

                    is_block = True
                    last_item_order = None
                    while is_block and selected_idx >= 0:
                        cur_item_order = selected_items.iloc[selected_idx][order_col_name]

                        if last_item_order is None or last_item_order - cur_item_order == 1:
                            current_block_exp_ids.append(selected_items.index[selected_idx])
                            selected_idx -= 1
                            last_item_order = cur_item_order
                        else:
                            is_block = False

                    order_of_highest_block_item = qgrid_df.loc[current_block_exp_ids[0]][order_col_name]

                    # only change orders if the block is not at the end of the list
                    if order_of_highest_block_item < len(qgrid_df) - 1:

                        # change the order of item below the block
                        exp_id_of_current_item_above_block = all_items_sorted_by_order.loc[order_of_highest_block_item + 1][experiment_id_col_name]

                        new_order = order_of_highest_block_item - len(current_block_exp_ids) + 1
                        self.qgrid_widget.edit_cell(
                            exp_id_of_current_item_above_block,
                            order_col_name,
                            new_order)
                        self.experiment_descriptions[exp_id_of_current_item_above_block]['order'] = new_order

                        # subtract 1 to all selected items in the block
                        for block_item_exp_id in current_block_exp_ids:
                            new_order = qgrid_df.loc[block_item_exp_id]['order'] + 1

                            self.qgrid_widget.edit_cell(
                                block_item_exp_id,
                                order_col_name,
                                new_order)
                            self.experiment_descriptions[block_item_exp_id]['order'] = new_order

                        #resort the experiments in the grid according to the order field
                        self.sort_grid_by_order()

                        # reslect the old elements
                        if is_select_changed_items:
                            self.qgrid_widget.change_selection(selected_items.index)

                        self.backup_state()

                        self._call_experiment_descriptions_updated_event()
        finally:
            self._handle_qgrid_cell_edited_is_active = True


    def move_experiments_down(self, selected_items=None, is_select_changed_items=True):

        try:
            self._handle_qgrid_cell_edited_is_active = False

            if selected_items is None:
                selected_items = self.qgrid_widget.get_selected_df()

            if len(selected_items) > 0:

                order_col_name = self.config.dataframe_column_names['order']
                experiment_id_col_name = self.config.dataframe_column_names['id']

                # select according to the order
                selected_items = selected_items.sort_values(by=order_col_name)
                qgrid_df = self.qgrid_widget.get_changed_df()

                # make order the index to allow to seach by it
                all_items_sorted_by_order = qgrid_df.sort_values(by=order_col_name).reset_index().set_index(order_col_name)

                selected_idx = 0
                while selected_idx < len(selected_items):

                    # identify current block of consequetively selected items
                    current_block_exp_ids = []

                    is_block = True
                    last_item_order = None
                    while is_block and selected_idx < len(selected_items):
                        cur_item_order = selected_items.iloc[selected_idx][order_col_name]

                        if last_item_order is None or cur_item_order - last_item_order == 1:
                            current_block_exp_ids.append(selected_items.index[selected_idx])
                            selected_idx += 1
                            last_item_order = cur_item_order
                        else:
                            is_block = False

                    order_of_lowest_block_item = qgrid_df.loc[current_block_exp_ids[0]][order_col_name]

                    # only change orders if the block is not at the end of the list
                    if order_of_lowest_block_item > 0:

                        # change the order of item below the block
                        exp_id_of_current_item_below_block = all_items_sorted_by_order.loc[order_of_lowest_block_item - 1][experiment_id_col_name]

                        new_order = order_of_lowest_block_item + len(current_block_exp_ids) - 1
                        self.qgrid_widget.edit_cell(
                            exp_id_of_current_item_below_block,
                            order_col_name,
                            new_order)
                        self.experiment_descriptions[exp_id_of_current_item_below_block]['order'] = new_order

                        # subtract 1 to all selected items in the block
                        for block_item_exp_id in current_block_exp_ids:
                            new_order = qgrid_df.loc[block_item_exp_id]['order'] - 1

                            self.qgrid_widget.edit_cell(
                                block_item_exp_id,
                                order_col_name,
                                new_order)
                            self.experiment_descriptions[block_item_exp_id]['order'] = new_order

                        # resort the experiments in the grid according to the order field
                        self.sort_grid_by_order()

                        # reselect the old elements
                        if is_select_changed_items:
                            self.qgrid_widget.change_selection(selected_items.index)

                        self.backup_state()

                        self._call_experiment_descriptions_updated_event()
        finally:
            self._handle_qgrid_cell_edited_is_active = True


    def resort_experiments_by_id(self):
        """
        Resets the order of the experiments according to their IDs.
        """
        try:
            # don't allow to change the qgrid during the operation
            self._handle_qgrid_cell_edited_is_active = False

            # sort experiments according to ids
            # get list of experiment ids
            sorted_existing_experiment_ids = sorted(list(self.experiment_descriptions.keys()))

            # update the order of the experiments according to the sorted list
            for order, exp_id in enumerate(sorted_existing_experiment_ids):
                self.experiment_descriptions[exp_id].order = order

            # update the gui according to the new order
            self._update_qgrid()

            # inform others the descriptions are changed
            self._call_experiment_descriptions_updated_event()

        finally:
            # qgrid can be changed again
            self._handle_qgrid_cell_edited_is_active = True


    def update_experiment_descriptions(self, is_reset=False):
        """Updates the experiment descriptions by adding new experiments and removing old experiments."""

        # load experiment descriptions
        new_exp_descr = eu.misc.call_function_from_config(
            self.config.load_experiment_descriptions_function,
            self.config.experiments_directory)

        if not self.experiment_descriptions or is_reset:
            self.experiment_descriptions = new_exp_descr
        else:
            # combine existing descriptions and new list

            # remove non-existing elements from exisiting descriptions
            deleted_experiments = set(self.experiment_descriptions.keys()).difference(set(new_exp_descr.keys()))
            for deleted_exp in deleted_experiments:
                del self.experiment_descriptions[deleted_exp]

            # kepp current order and add new experiments at the end of the list
            # get current order of experiment
            sorted_existing_experiment_ids = eu.data.get_ordered_experiment_ids_from_descriptions(self.experiment_descriptions)
            sorted_new_experiment_ids = eu.data.get_ordered_experiment_ids_from_descriptions(new_exp_descr)

            # remove existing experiment ids from the sorted list of new experiment ids
            for existing_exp_id in sorted_existing_experiment_ids:
                if existing_exp_id in sorted_new_experiment_ids:
                    sorted_new_experiment_ids.remove(existing_exp_id)

            # add new elements
            self.experiment_descriptions = eu.combine_dicts(self.experiment_descriptions, new_exp_descr)

            # update the order of the experiments according to the sorted lists
            for order, exp_id in enumerate(sorted_existing_experiment_ids + sorted_new_experiment_ids):
                self.experiment_descriptions[exp_id].order = order

            # do not keep the repetition ids from existing ones, but use the ones from the new discriptions
            # otherwise, if new repetitions are added, they will not be used
            for new_descr in new_exp_descr.values():
                self.experiment_descriptions[new_descr.id].repetition_ids = new_descr.repetition_ids

        self._call_experiment_descriptions_updated_event()


    def get_widget_state(self):
        state = super().get_widget_state()
        state.experiment_descriptions = self.experiment_descriptions
        return state


    def set_widget_state(self, state):
        if 'experiment_descriptions' in state: self.experiment_descriptions  = state.experiment_descriptions
        return super().set_widget_state(state)

    def empty_data(self):
        """Delete experiment_data to free memory."""
        if self.experiment_data:
            keys = list(self.experiment_data.keys())
            for key in keys:
                del self.experiment_data[key]

    def load_data(self):

        # delete old data to free memory
        self.empty_data()

        experiment_data = eu.misc.call_function_from_config(
            self.config.load_experiment_data_function,
            self.experiment_descriptions)

        # some data loader functions give as extra argument the experiment descriptions
        if isinstance(experiment_data, tuple):
            experiment_data = experiment_data[0]

        self.experiment_data = experiment_data

        self._call_experiment_data_loaded_event()



##
## This file is part of the exputils package.
##
## Copyright: INRIA
## Year: 2022
## Contact: chris.reinke@inria.fr
##
## exputils is provided under GPL-3.0-or-later
##
import exputils as eu
import numpy as np
import base64
import cloudpickle
try:
    import json
except ImportError:
    import simplejson as json


class ExputilsJSONEncoder(json.JSONEncoder):
    """
    Encodes objects into json strings. Allows to encode numpy arrays and functions.
    The exputils_json_object_hook() function can be used to load the dumped objects.
    Usage:
    >>> dumped = json.dumps(object, cls=ExputilsJSONEncoder)
    >>> loaded = json.loads(dumped, object_hook=exputils_json_object_hook)
    """

    def default(self, obj):
        """
        if input object is a ndarray it will be converted into a dict holding dtype, shape and the data base64 encoded
        """
        if isinstance(obj, np.ndarray):
            data_b64 = base64.b64encode(np.ascontiguousarray(obj).data)
            return dict(__ndarray__=data_b64.decode('ascii'),
                        dtype=str(obj.dtype),
                        shape=obj.shape)
        # Let the base class default method raise the TypeError

        if isinstance(obj, np.int64):
            return int(obj)

        if isinstance(obj, np.float64):
            return float(obj)

        if callable(obj):
            b = cloudpickle.dumps(obj)
            s = base64.b64encode(b).decode()
            return dict(__cloudpickle__=s)

        return super().default(obj)


def exputils_json_object_hook(dct):
    """
    Decodes json data. Allows to decode numpy arrays and functions.

    Usage:
    >>> dumped = json.dumps(object, cls=ExputilsJSONEncoder)
    >>> loaded = json.loads(dumped, object_hook=exputils_json_object_hook)
    """
    if isinstance(dct, dict) and '__ndarray__' in dct:
        data = base64.b64decode(dct['__ndarray__'])
        return np.frombuffer(data, dct['dtype']).reshape(dct['shape'])

    if isinstance(dct, dict) and '__cloudpickle__' in dct:
        b = base64.b64decode(dct['__cloudpickle__'])
        expr = cloudpickle.loads(b)
        return expr

    return dct


def save_dict_as_json_file(dict, filepath, **kwargs):
    """ Saves a dict to a JSON file. Accepts the same keyword options as `json.dumps()`."""

    # allow saving of numpy arrays in JSON file
    if 'cls' not in kwargs:
        kwargs['cls'] = eu.io.json.ExputilsJSONEncoder

    json_content = json.dumps(dict, **kwargs)

    eu.io.makedirs_for_file(filepath)

    f = open(filepath, "w")
    f.write(json_content)
    f.close()


def load_dict_from_json_file(filepath, is_transform_ints=False, **kwargs):
    """
    Loads a dictionary from a JSON file.

    :param filepath: Path to the file.
    :param is_transform_ints: JSON does not allow to store integers a dictionary keus.
                              If this option is set to True any keys in the JSON are tried to conver to integers if possible.
    :param kwargs: Optional arguments that are passed to the json.loads function.
    :return: Loaded dictionary.
    """

    if 'cls' not in kwargs:
        kwargs['object_hook'] = eu.io.json.exputils_json_object_hook

    # Open the file in read mode
    file_object = open(filepath, 'r')
    json_content = file_object.read()
    file_object.close()

    loaded_dict = json.loads(json_content, **kwargs)

    # chenge possible int keys to ints as json changes ints to strings
    if is_transform_ints:
        loaded_dict = convert_json_dict_keys_to_ints(loaded_dict)

    return loaded_dict


def convert_json_dict_keys_to_ints(json_data):
    '''
    JSON does not allow ints as keys for dictionaries. This method transforms all keys in the given json data to ints if such a conversion is possible.

    :param json_data:
    :return:
    '''

    correctedDict = {}

    for key, value in json_data.items():
        if isinstance(value, list):
            value = [convert_json_dict_keys_to_ints(item) if isinstance(item, dict) else item for item in value]
        elif isinstance(value, dict):
            value = convert_json_dict_keys_to_ints(value)
        try:
            key = int(key)
        except Exception as ex:
            pass
        correctedDict[key] = value

    return correctedDict




##
## This file is part of the exputils package.
##
## Copyright: INRIA
## Year: 2022
## Contact: chris.reinke@inria.fr
##
## exputils is provided under GPL-3.0-or-later
##
import exputils as eu
from exputils.data.logger import Logger

# holds the global logger object
log = Logger()


def reset():
    """
    Resets the log to an empty exputils.data.Logger.
    """
    global log
    log = Logger()


def get_log():
    """
    Returns the current logger.
    """
    global log
    return log


def set_log(new_log):
    """
    Sets the given logger to be the global log
    """
    global log
    log = new_log


def set_directory(directory):
    """
    Sets the directory path for the log.
    """
    log.directory = directory


def get_directory():
    """
    Returns the directory path of the log.
    """
    return log.directory


def contains(name):
    """
    Returns True if items for the the given name exists in the log. Otherwise False.
    """
    return (name in log)


def clear(name=None):
    """
    Clears the data of the whole log or of a specific data element.

    :param name: If none, then the whole log is cleared, otherwise only the data element with the given name.
                 (default=None)
    """
    log.clear(name=name)


def get_item(name):
    """
    Returns the item from the log with the given name.
    """
    return log[name]


def add_value(name, value, log_to_tb=None, tb_global_step=None, tb_walltime=None):
    """
    Adds a new value to the log. Values are stored in numpy arrays. Allows to log in parallel to tensorboard if the value is a scalar.

    :param: name (string): Name of the value. (Can use dividers '/' for tensorbard. They are replaced by '_' for the normal log. )
    :param: scalar: Value to save.
    :param log_to_tb (bool): True if the value should be logged to tensorboard. False if not.
        If None, then it gets logged if the tensorboard is active. (default = None)
    :param tb_global_step (int): Tensorboards global step value to record. (default = None)
    :param tb_walltime (float): Optional tensorboard override of default walltime (time.time()) with seconds after epoch of event. (default = None)
    """
    log.add_value(name, value, log_to_tb, tb_global_step, tb_walltime)


def add_scalar(name, scalar, log_to_tb=None, tb_global_step=None, tb_walltime=None):
    """
    Adds a new scalar to the log. Scalars are stored in numpy arrays. Allows to log in parallel to tensorboard.

    :param: name (string): Name of the value. (Can use dividers '/' for tensorbard. They are replaced by '_' for the normal log. )
    :param: scalar: Value to save.
    :param log_to_tb (bool): True if the value should be logged to tensorboard. False if not.
        If None, then it gets logged if the tensorboard is active. (default = None)
    :param tb_global_step (int): Tensorboards global step value to record. (default = None)
    :param tb_walltime (float): Optional tensorboard override of default walltime (time.time()) with seconds after epoch of event. (default = None)
    """
    log.add_scalar(name, scalar, log_to_tb, tb_global_step, tb_walltime)


def get_values(name):
    """
    Returns the values for the given name. Values are stored in numpy arrays.
    """
    return log[name]


def add_object(name, obj):
    """
    Adds a new object to the log. Objects are stored in a list and saved as files using dill.
    """
    log.add_object(name, obj)


def get_objects(name):
    """
    Returns the objects for the given name. Objects are stored in a list.
    """
    return log[name]


def add_single_object(name, obj, directory=None):
    """
    Adds a single object to the log which is directly written to the hard drive and not stored in memory.
    The objects is saved via dill.
    """
    log.add_single_object(name, obj, directory=directory)


def items():
    """
    Returns the items in the log as a list of tuples with the name and values of the items.
    """
    return log.items()


def save(directory=None):
    """
    Saves the log to its defined directory.

    :param directory: Optional path to the directory.
    """
    log.save(directory=directory)


def load(directory=None, load_objects=False):
    """
    Loads the items from a log directory into the log.

    :param directory: Optional path to the directory.
    :param load_objects: If True then also objects (dill files) are loaded. Default: False.
    """
    log.load(directory=directory, load_objects=load_objects)


def load_single_object(name):
    """
    Loads a single object from a log and returns it.

    :return:
    """
    return log.load_single_object(name)


def set_config(config=None, **kwargs):
    """
    Sets the config of the log.

    :param config: Dictionary with config parameters.
    :param kwargs: Arguments list of config parameters.
    """
    log.config = eu.combine_dicts(kwargs, config, log.config)


####################################################
# TENSORBOARD

def tensorboard():
    """The tensorboard SummaryWriter that can be used to log data to the tensorboard.

    The logged data will only be stored in the tensorboard logs!

    Example:

    log.

    """
    return log.tensorboard


def create_tensorboard(config=None, **kwargs):
    """
    Creates a tensorboard that can be used for logging.

    :param config (dict): Dictionary with the configuration of the tensorboard. Has the same entries as the parameters below. (default = None)
    :param log_dir (string): Save directory location. Default is experiments/tensorboard_logs/exp_<experiment_id>/rep_<repetition_id>/<date>_<time>.
    :param purge_step (int): When logging crashes at step T+XT+XT+X and restarts at step TTT, any events whose global_step larger or equal to TTT
        will be purged and hidden from TensorBoard. Note that crashed and resumed experiments should have the same log_dir.
    :param: max_queue (int): Size of the queue for pending events and summaries before one of the ‘add’ calls forces a flush to disk. (default = 10)
    :param flush_secs (int): How often, in seconds, to flush the pending events and summaries to disk. (default = 120)
    :param filename_suffix (string): Suffix added to all event filenames in the log_dir directory. (default = '.tblog')

    :return: SummaryWriter of the tensorboard (See https://pytorch.org/docs/stable/tensorboard.html).
    """

    return log.create_tensorboard(config=config, **kwargs)


def activate_tensorboard(config=None, **kwargs):
    """
    Activates a tensorboard that can be used for logging.
    If it is activated, then when the function add_value/add_scalar is used and a scalar is given, the tensorboard automatically logs it too.
    Creates a tensorboard if non existed so far.

    :param config (dict): Dictionary with the configuration of the tensorboard. Has the same entries as the parameters below. (default = None)
    :param log_dir (string): Save directory location. Default is experiments/tensorboard_logs/exp_<experiment_id>/rep_<repetition_id>/<date>_<time>.
    :param purge_step (int): When logging crashes at step T+XT+XT+X and restarts at step TTT, any events whose global_step larger or equal to TTT
        will be purged and hidden from TensorBoard. Note that crashed and resumed experiments should have the same log_dir.
    :param: max_queue (int): Size of the queue for pending events and summaries before one of the ‘add’ calls forces a flush to disk. (default = 10)
    :param flush_secs (int): How often, in seconds, to flush the pending events and summaries to disk. (default = 120)
    :param filename_suffix (string): Suffix added to all event filenames in the log_dir directory. (default = '.tblog')

    :return: SummaryWriter of the tensorboard (See https://pytorch.org/docs/stable/tensorboard.html).
    """

    return log.activate_tensorboard(config=config, **kwargs)


def deactivate_tensorboard():
    """
    Deactivates a tensorboard. Afterwards, values will not be automatically logged via the add_value / add_scalar function to the tensorboard.
    """
    return log.deactivate_tensorboard()


def is_tensorboard_active():
    """Returns true, if the tensorboard is active."""
    return log.is_tensorboard_active






##
## This file is part of the exputils package.
##
## Copyright: INRIA
## Year: 2022
## Contact: chris.reinke@inria.fr
##
## exputils is provided under GPL-3.0-or-later
##
# TODO: Feature - Saving and loading of pytorch models